# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import mlrun.common.runtimes.constants as runtimes_constants
import mlrun.common.schemas
import mlrun.lists
import mlrun.utils.helpers

from .base import NotificationBase, TimedHTTPClient


class SlackNotification(NotificationBase):
    """
    API/Client notification for sending run statuses as a slack message
    """

    emojis = {
        "completed": ":smiley:",
        "running": ":man-running:",
        "error": ":x:",
        "skipped": ":zzz:",
    }

    @classmethod
    def validate_params(cls, params):
        webhook = params.get("webhook", None) or mlrun.get_secret_or_env(
            "SLACK_WEBHOOK"
        )
        if not webhook:
            raise ValueError("Parameter 'webhook' is required for SlackNotification")

    async def push(
        self,
        message: str,
        severity: typing.Optional[
            typing.Union[mlrun.common.schemas.NotificationSeverity, str]
        ] = mlrun.common.schemas.NotificationSeverity.INFO,
        runs: typing.Optional[typing.Union[mlrun.lists.RunList, list]] = None,
        custom_html: typing.Optional[typing.Optional[str]] = None,
        alert: typing.Optional[mlrun.common.schemas.AlertConfig] = None,
        event_data: typing.Optional[mlrun.common.schemas.Event] = None,
    ):
        webhook = self.params.get("webhook", None) or mlrun.get_secret_or_env(
            "SLACK_WEBHOOK"
        )

        if not webhook:
            mlrun.utils.helpers.logger.debug(
                "No slack webhook is set, skipping notification"
            )
            return

        data = self._generate_slack_data(message, severity, runs, alert, event_data)

        async with TimedHTTPClient().session() as session:
            async with session.post(webhook, json=data) as response:
                response.raise_for_status()

    def _generate_slack_data(
        self,
        message: str,
        severity: typing.Union[
            mlrun.common.schemas.NotificationSeverity, str
        ] = mlrun.common.schemas.NotificationSeverity.INFO,
        runs: typing.Union[mlrun.lists.RunList, list] = None,
        alert: mlrun.common.schemas.AlertConfig = None,
        event_data: mlrun.common.schemas.Event = None,
    ) -> dict:
        data = {
            "blocks": self._generate_slack_header_blocks(severity, message),
        }
        if self.name:
            data["blocks"].append(
                {"type": "section", "text": self._get_slack_row(self.name)}
            )

        if alert:
            fields = self._get_alert_fields(alert, event_data)

            for i in range(len(fields)):
                data["blocks"].append({"type": "section", "text": fields[i]})
        else:
            if not runs:
                return data

            if isinstance(runs, list):
                runs = mlrun.lists.RunList(runs)

            fields = [self._get_slack_row("*Runs*"), self._get_slack_row("*Results*")]
            for run in runs:
                fields.append(self._get_run_line(run))
                fields.append(self._get_run_result(run))

            for i in range(0, len(fields), 8):
                data["blocks"].append({"type": "section", "fields": fields[i : i + 8]})

        return data

    def _generate_slack_header_blocks(self, severity: str, message: str):
        header_text = block_text = f"[{severity}] {message}"
        section_text = None

        # Slack doesn't allow headers to be longer than 150 characters
        # If there's a comma in the message, split the message at the comma
        # Otherwise, split the message at 150 characters
        if len(block_text) > 150:
            if ", " in block_text and block_text.index(", ") < 149:
                header_text = block_text.split(",")[0]
                section_text = block_text[len(header_text) + 2 :]
            else:
                header_text = block_text[:150]
                section_text = block_text[150:]
        blocks = [
            {"type": "header", "text": {"type": "plain_text", "text": header_text}}
        ]
        if section_text:
            blocks.append(
                {
                    "type": "section",
                    "text": self._get_slack_row(section_text),
                }
            )
        return blocks

    def _get_alert_fields(
        self,
        alert: mlrun.common.schemas.AlertConfig,
        event_data: mlrun.common.schemas.Event,
    ) -> list:
        line = [
            self._get_slack_row(f":bell: {alert.name} alert has occurred"),
            self._get_slack_row(f"*Project:*\n{alert.project}"),
            self._get_slack_row(f"*ID:*\n{event_data.entity.ids[0]}"),
        ]

        if alert.summary:
            line.append(
                self._get_slack_row(
                    f"*Summary:*\n{mlrun.utils.helpers.format_alert_summary(alert, event_data)}"
                )
            )

        if event_data.value_dict:
            data_lines = []
            for key, value in event_data.value_dict.items():
                data_lines.append(f"{key}: {value}")
            data_text = "\n".join(data_lines)
            line.append(self._get_slack_row(f"*Event data:*\n{data_text}"))

        overview_type, url = self._get_overview_type_and_url(alert, event_data)
        line.append(self._get_slack_row(f"*Overview:*\n<{url}|*{overview_type}*>"))

        return line

    def _get_run_line(self, run: dict) -> dict:
        meta = run["metadata"]
        url = mlrun.utils.helpers.get_run_url(
            meta.get("project"),
            uid=meta.get("uid"),
            name=meta.get("name"),
        )

        # Only show the URL if the run is not a function (serving or mlrun function)
        kind = run.get("step_kind")
        state = run["status"].get("state", "")

        if state != runtimes_constants.RunStates.skipped and (
            url and not kind or kind == "run"
        ):
            line = f'<{url}|*{meta.get("name")}*>'
        else:
            line = meta.get("name")
        if kind:
            line = f'{line} *({run.get("step_kind", run.get("kind", ""))})*'
        line = f'{self.emojis.get(state, ":question:")}  {line}'
        return self._get_slack_row(line)

    def _get_run_result(self, run: dict) -> dict:
        state = run["status"].get("state", "")
        if state == "error":
            error_status = run["status"].get("error", "") or state
            result = f"*{error_status}*"
        else:
            result = mlrun.utils.helpers.dict_to_str(
                run["status"].get("results", {}), ", "
            )
        return self._get_slack_row(result or state)

    @staticmethod
    def _get_slack_row(text: str) -> dict:
        return {"type": "mrkdwn", "text": text}



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import sys
from importlib.resources import read_text

import mlrun.utils
from mlrun.utils.singleton import Singleton


class _VersionInfo:
    def __init__(self, major, minor, patch):
        self.major = major
        self.minor = minor
        self.patch = patch

    def __str__(self):
        return f"{self.major}.{self.minor}.{self.patch}"


class Version(metaclass=Singleton):
    def __init__(self):
        # When installing un-released version (e.g. by doing pip install git+https://github.com/mlrun/mlrun@development)
        # it won't have a version file, so adding some sane defaults
        self.version_info = {"git_commit": "unknown", "version": "0.0.0+unstable"}
        self.python_version = self._resolve_python_version()
        try:
            self.version_info = json.loads(
                read_text("mlrun.utils.version", "version.json")
            )
        except Exception:
            mlrun.utils.logger.warning(
                "Failed resolving version info. Ignoring and using defaults"
            )

    def get(self):
        return self.version_info

    def get_python_version(self) -> _VersionInfo:
        return self.python_version

    @staticmethod
    def _resolve_python_version() -> _VersionInfo:
        return _VersionInfo(*sys.version_info[:3])



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import importlib.util
import pathlib
import sys
from datetime import datetime
from typing import Any, Optional, Union

import pandas as pd

import mlrun
import mlrun.errors

from ..data_types import InferOptions, get_infer_interface
from ..datastore.sources import BaseSourceDriver, StreamSource
from ..datastore.store_resources import parse_store_uri
from ..datastore.targets import (
    BaseStoreTarget,
    get_default_prefix_for_source,
    get_target_driver,
    kind_to_driver,
    validate_target_list,
    validate_target_paths_for_engine,
    write_spark_dataframe_with_options,
)
from ..model import DataSource, DataTargetBase
from ..runtimes import BaseRuntime, RuntimeKinds
from ..runtimes.function_reference import FunctionReference
from ..serving.server import Response
from ..utils import get_caller_globals, logger, normalize_name
from .common import (
    RunConfig,
    get_feature_set_by_uri,
    get_feature_vector_by_uri,
    verify_feature_set_exists,
    verify_feature_set_permissions,
)
from .feature_set import FeatureSet
from .ingestion import (
    context_to_ingestion_params,
    init_featureset_graph,
    run_ingestion_job,
    run_spark_graph,
)

_v3iofs = None
spark_transform_handler = "transform"
_TRANS_TABLE = str.maketrans({" ": "_", "(": "", ")": ""})


def norm_column_name(name: str) -> str:
    """
    Remove parentheses () and replace whitespaces with an underscore _.
    Used to normalize a column/feature name.
    """
    return name.translate(_TRANS_TABLE)


def _rename_source_dataframe_columns(df: pd.DataFrame) -> pd.DataFrame:
    rename_mapping = {}
    column_set = set(df.columns)
    for column in df.columns:
        if isinstance(column, str):
            rename_to = norm_column_name(column)
            if rename_to != column:
                if rename_to in column_set:
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        f'column "{column}" cannot be renamed to "{rename_to}" because such a column already exists'
                    )
                rename_mapping[column] = rename_to
                column_set.add(rename_to)
    if rename_mapping:
        logger.warn(
            f"the following dataframe columns have been renamed due to unsupported characters: {rename_mapping}"
        )
        df = df.rename(rename_mapping, axis=1)
    return df


def _get_namespace(run_config: RunConfig) -> dict[str, Any]:
    # if running locally, we need to import the file dynamically to get its namespace
    if run_config and run_config.local and run_config.function:
        filename = run_config.function.spec.filename
        if filename:
            module_name = pathlib.Path(filename).name.rsplit(".", maxsplit=1)[0]
            spec = importlib.util.spec_from_file_location(module_name, filename)
            module = importlib.util.module_from_spec(spec)
            sys.modules[module_name] = module
            spec.loader.exec_module(module)
            return vars(__import__(module_name))
    else:
        return get_caller_globals()


def ingest(
    mlrun_context: Union["mlrun.MLrunProject", "mlrun.MLClientCtx"],
    featureset: Union[FeatureSet, str] = None,
    source=None,
    targets: Optional[list[DataTargetBase]] = None,
    namespace=None,
    return_df: bool = True,
    infer_options: InferOptions = InferOptions.default(),
    run_config: RunConfig = None,
    spark_context=None,
    overwrite=None,
) -> Optional[pd.DataFrame]:
    """Read local DataFrame, file, URL, or source into the feature store
    Ingest reads from the source, run the graph transformations, infers  metadata and stats
    and writes the results to the default of specified targets

    when targets are not specified data is stored in the configured default targets
    (will usually be NoSQL for real-time and Parquet for offline).

    the `run_config` parameter allow specifying the function and job configuration,
    see: :py:class:`~mlrun.feature_store.RunConfig`

    example::

        stocks_set = FeatureSet("stocks", entities=[Entity("ticker")])
        stocks = pd.read_csv("stocks.csv")
        df = ingest(stocks_set, stocks, infer_options=fstore.InferOptions.default())

        # for running as remote job
        config = RunConfig(image="mlrun/mlrun")
        df = ingest(stocks_set, stocks, run_config=config)

        # specify source and targets
        source = CSVSource("mycsv", path="measurements.csv")
        targets = [CSVTarget("mycsv", path="./mycsv.csv")]
        ingest(measurements, source, targets)

    :param mlrun_context: mlrun context
    :param featureset:    feature set object or featureset.uri. (uri must be of a feature set that is in the DB,
                          call `.save()` if it's not)
    :param source:        source dataframe or other sources (e.g. parquet source see:
                          :py:class:`~mlrun.datastore.ParquetSource` and other classes in mlrun.datastore with suffix
                          Source)
    :param targets:       optional list of data target objects
    :param namespace:     namespace or module containing graph classes
    :param return_df:     indicate if to return a dataframe with the graph results
    :param infer_options: schema (for discovery of entities, features in featureset), index, stats,
                          histogram and preview infer options (:py:class:`~mlrun.feature_store.InferOptions`)
    :param run_config:    function and/or run configuration for remote jobs,
                          see :py:class:`~mlrun.feature_store.RunConfig`
    :param spark_context: local spark session for spark ingestion, example for creating the spark context:
                          `spark = SparkSession.builder.appName("Spark function").getOrCreate()`
                          For remote spark ingestion, this should contain the remote spark service name
    :param overwrite:     delete the targets' data prior to ingestion
                          (default: True for non scheduled ingest - deletes the targets that are about to be ingested.
                          False for scheduled ingest - does not delete the target)
    :return:              if return_df is True, a dataframe will be returned based on the graph
    """
    if not mlrun_context:
        raise mlrun.errors.MLRunValueError(
            "mlrun_context must be defined when calling ingest()"
        )

    return _ingest(
        featureset,
        source,
        targets,
        namespace,
        return_df,
        infer_options,
        run_config,
        mlrun_context,
        spark_context,
        overwrite,
    )


def _ingest(
    featureset: Union[FeatureSet, str] = None,
    source=None,
    targets: Optional[list[DataTargetBase]] = None,
    namespace=None,
    return_df: bool = True,
    infer_options: InferOptions = InferOptions.default(),
    run_config: RunConfig = None,
    mlrun_context=None,
    spark_context=None,
    overwrite=None,
) -> Optional[pd.DataFrame]:
    if isinstance(source, pd.DataFrame):
        source = _rename_source_dataframe_columns(source)

    if featureset:
        if isinstance(featureset, str):
            # need to strip store prefix from the uri
            _, stripped_name = parse_store_uri(featureset)
            try:
                featureset = get_feature_set_by_uri(stripped_name)
            except mlrun.db.RunDBError as exc:
                # TODO: this handling is needed because the generic httpdb error handling doesn't raise the correct
                #  error class and doesn't propagate the correct message, until it solved we're manually handling this
                #  case to give better user experience, remove this when the error handling is fixed.
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"{exc}. Make sure the feature set is saved in DB (call feature_set.save())"
                )

        # feature-set spec always has a source property that is not None. It may be default-constructed, in which
        # case the path will be 'None'. That's why we need a special check
        if source is None and featureset.has_valid_source():
            source = featureset.spec.source

    if not mlrun_context and (not featureset or source is None):
        raise mlrun.errors.MLRunInvalidArgumentError(
            "feature set and source must be specified"
        )
    if (
        not mlrun_context
        and not targets
        and not (featureset.spec.targets or featureset.spec.with_default_targets)
        and (run_config is not None and not run_config.local)
    ):
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"Feature set {featureset.metadata.name} is remote ingested with no targets defined, aborting"
        )

    if featureset is not None:
        featureset.validate_steps(namespace=namespace)
    # This flow may happen both on client side (user provides run config) and server side (through the ingest API)
    if run_config and not run_config.local:
        if isinstance(source, pd.DataFrame):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "DataFrame source is illegal in conjunction with run_config"
            )
        # remote job execution
        verify_feature_set_permissions(
            featureset, mlrun.common.schemas.AuthorizationAction.update
        )
        run_config = run_config.copy() if run_config else RunConfig()
        source, run_config.parameters = set_task_params(
            featureset, source, targets, run_config.parameters, infer_options, overwrite
        )
        name = f"{featureset.metadata.name}_ingest"
        schedule = source.schedule
        if schedule == "mock":
            schedule = None
        return run_ingestion_job(name, featureset, run_config, schedule, spark_context)

    if mlrun_context:
        # extract ingestion parameters from mlrun context
        if isinstance(source, pd.DataFrame):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "DataFrame source is illegal when running ingest remotely"
            )
        if featureset or source is not None:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "cannot specify mlrun_context with feature set or source"
            )
        (
            featureset,
            source,
            targets,
            infer_options,
            overwrite,
        ) = context_to_ingestion_params(mlrun_context)

        featureset.validate_steps(namespace=namespace)
        verify_feature_set_permissions(
            featureset, mlrun.common.schemas.AuthorizationAction.update
        )
        if not source:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "data source was not specified"
            )

        filter_time_string = ""
        if source.schedule:
            featureset.reload(update_spec=False)

    if isinstance(source, DataSource) and source.schedule:
        min_time = datetime.max
        for target in featureset.status.targets:
            if target.last_written:
                cur_last_written = target.last_written
                if isinstance(cur_last_written, str):
                    cur_last_written = datetime.fromisoformat(target.last_written)
                if cur_last_written < min_time:
                    min_time = cur_last_written
        if min_time != datetime.max:
            source.start_time = min_time
            time_zone = min_time.tzinfo
            source.end_time = datetime.now(tz=time_zone)
            filter_time_string = (
                f"Source.start_time for the job is{str(source.start_time)}. "
                f"Source.end_time is {str(source.end_time)}"
            )

        if mlrun_context:
            mlrun_context.logger.info(
                f"starting ingestion task to {featureset.uri}.{filter_time_string}"
            )

        return_df = False

    if featureset.spec.passthrough:
        featureset.spec.source = source
        featureset.spec.validate_no_processing_for_passthrough()

    if not namespace:
        namespace = _get_namespace(run_config)

    targets_to_ingest = targets or featureset.spec.targets
    targets_to_ingest = copy.deepcopy(targets_to_ingest)

    validate_target_paths_for_engine(targets_to_ingest, featureset.spec.engine, source)

    if overwrite is None:
        if isinstance(source, BaseSourceDriver) and source.schedule:
            overwrite = False
        else:
            overwrite = True

    if overwrite:
        validate_target_list(targets=targets_to_ingest)
        purge_target_names = [
            t if isinstance(t, str) else t.name for t in targets_to_ingest
        ]
        featureset.purge_targets(target_names=purge_target_names, silent=True)

        featureset.update_targets_for_ingest(
            targets=targets_to_ingest,
            overwrite=overwrite,
        )
    else:
        featureset.update_targets_for_ingest(
            targets=targets_to_ingest,
            overwrite=overwrite,
        )

        for target in targets_to_ingest:
            if not kind_to_driver[target.kind].support_append:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"{target.kind} target does not support overwrite=False ingestion"
                )
            if hasattr(target, "is_single_file") and target.is_single_file():
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "overwrite=False isn't supported in single files. Please use folder path."
                )

    if spark_context and featureset.spec.engine != "spark":
        raise mlrun.errors.MLRunInvalidArgumentError(
            "featureset.spec.engine must be set to 'spark' to ingest with spark"
        )
    if featureset.spec.engine == "spark":
        import pyspark.sql

        if (
            isinstance(source, (pd.DataFrame, pyspark.sql.DataFrame))
            and run_config is not None
        ):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "DataFrame source is illegal when ingesting with remote spark or spark operator"
            )
        # use local spark session to ingest
        return _ingest_with_spark(
            spark_context,
            featureset,
            source,
            targets_to_ingest,
            infer_options=infer_options,
            mlrun_context=mlrun_context,
            namespace=namespace,
            overwrite=overwrite,
            return_df=return_df,
        )

    if isinstance(source, str):
        source = mlrun.store_manager.object(url=source).as_df()

    schema_options = InferOptions.get_common_options(
        infer_options, InferOptions.schema()
    )
    if schema_options:
        _preview(
            featureset,
            source,
            options=schema_options,
            namespace=namespace,
        )
    infer_stats = InferOptions.get_common_options(
        infer_options, InferOptions.all_stats()
    )
    # Check if dataframe is already calculated (for feature set graph):
    calculate_df = return_df or infer_stats != InferOptions.Null
    featureset.save()

    df = init_featureset_graph(
        source,
        featureset,
        namespace,
        targets=targets_to_ingest,
        return_df=calculate_df,
    )
    if not InferOptions.get_common_options(
        infer_stats, InferOptions.Index
    ) and InferOptions.get_common_options(infer_options, InferOptions.Index):
        infer_stats += InferOptions.Index

    _infer_from_static_df(df, featureset, options=infer_stats)

    if isinstance(source, DataSource):
        for target in featureset.status.targets:
            if (
                target.last_written == datetime.min
                and source.schedule
                and source.start_time
            ):
                # datetime.min is a special case that indicated that nothing was written in storey. we need the fix so
                # in the next scheduled run, we will have the same start time
                target.last_written = source.start_time

    _post_ingestion(mlrun_context, featureset, spark_context)
    if return_df:
        return df


def _preview(
    featureset: FeatureSet,
    source,
    entity_columns: Optional[list] = None,
    namespace=None,
    options: InferOptions = None,
    verbose: bool = False,
    sample_size: Optional[int] = None,
) -> pd.DataFrame:
    if isinstance(source, pd.DataFrame):
        source = _rename_source_dataframe_columns(source)

    # preview reads the source as a pandas df, which is not fully compatible with spark
    if featureset.spec.engine == "spark":
        raise mlrun.errors.MLRunInvalidArgumentError(
            "preview with spark engine is not supported"
        )

    options = options if options is not None else InferOptions.default()

    if isinstance(source, str):
        # if source is a path/url convert to DataFrame
        source = mlrun.store_manager.object(url=source).as_df()

    verify_feature_set_permissions(
        featureset, mlrun.common.schemas.AuthorizationAction.update
    )

    featureset.spec.validate_no_processing_for_passthrough()
    featureset.validate_steps(namespace=namespace)

    namespace = namespace or get_caller_globals()
    if featureset.spec.require_processing():
        _, default_final_step, _ = featureset.graph.check_and_process_graph(
            allow_empty=True
        )
        if not default_final_step:
            raise mlrun.errors.MLRunPreconditionFailedError(
                "Split flow graph must have a default final step defined"
            )
        # find/update entities schema
        if len(featureset.spec.entities) == 0:
            _infer_from_static_df(
                source,
                featureset,
                entity_columns,
                InferOptions.get_common_options(options, InferOptions.Entities),
            )
        # reduce the size of the ingestion if we do not infer stats
        rows_limit = (
            None
            if InferOptions.get_common_options(options, InferOptions.Stats)
            else 1000
        )
        source = init_featureset_graph(
            source,
            featureset,
            namespace,
            return_df=True,
            verbose=verbose,
            rows_limit=rows_limit,
        )

    df = _infer_from_static_df(
        source, featureset, entity_columns, options, sample_size=sample_size
    )
    featureset.save()
    return df


def _run_ingestion_job(
    featureset: Union[FeatureSet, str],
    source: DataSource = None,
    targets: Optional[list[DataTargetBase]] = None,
    name: Optional[str] = None,
    infer_options: InferOptions = InferOptions.default(),
    run_config: RunConfig = None,
):
    if isinstance(featureset, str):
        featureset = get_feature_set_by_uri(featureset)

    run_config = run_config.copy() if run_config else RunConfig()
    source, run_config.parameters = set_task_params(
        featureset, source, targets, run_config.parameters, infer_options
    )

    return run_ingestion_job(name, featureset, run_config, source.schedule)


def _deploy_ingestion_service_v2(
    featureset: Union[FeatureSet, str],
    source: DataSource = None,
    targets: Optional[list[DataTargetBase]] = None,
    name: Optional[str] = None,
    run_config: RunConfig = None,
    verbose=False,
) -> tuple[str, BaseRuntime]:
    if isinstance(featureset, str):
        featureset = get_feature_set_by_uri(featureset)

    verify_feature_set_permissions(
        featureset, mlrun.common.schemas.AuthorizationAction.update
    )

    verify_feature_set_exists(featureset)

    run_config = run_config.copy() if run_config else RunConfig()
    if isinstance(source, StreamSource) and not source.path:
        source.path = get_default_prefix_for_source(source.kind).format(
            project=featureset.metadata.project,
            kind=source.kind,
            name=featureset.metadata.name,
        )

    targets_to_ingest = targets or featureset.spec.targets
    targets_to_ingest = copy.deepcopy(targets_to_ingest)
    featureset.update_targets_for_ingest(targets_to_ingest)

    source, run_config.parameters = set_task_params(
        featureset, source, targets_to_ingest, run_config.parameters
    )

    name = normalize_name(name or f"{featureset.metadata.name}-ingest")
    if not run_config.function:
        function_ref = featureset.spec.function.copy()
        if function_ref.is_empty():
            function_ref = FunctionReference(name=name, kind=RuntimeKinds.serving)
        function_ref.kind = function_ref.kind or RuntimeKinds.serving
        if not function_ref.url:
            function_ref.code = function_ref.code or ""
        run_config.function = function_ref

    function = run_config.to_function(
        RuntimeKinds.serving, mlrun.mlconf.feature_store.default_job_image
    )
    function.metadata.project = featureset.metadata.project
    function.metadata.name = function.metadata.name or name

    function.spec.graph = featureset.spec.graph
    function.spec.graph.engine = (
        "async" if featureset.spec.engine == "storey" else "sync"
    )
    function.spec.parameters = run_config.parameters
    function.spec.graph_initializer = (
        "mlrun.feature_store.ingestion.featureset_initializer"
    )
    function.verbose = function.verbose or verbose
    function = source.add_nuclio_trigger(function)

    if run_config.local:
        return function.to_mock_server(namespace=get_caller_globals())
    return function.deploy(), function


def _ingest_with_spark(
    spark=None,
    featureset: Union[FeatureSet, str] = None,
    source: BaseSourceDriver = None,
    targets: Optional[list[BaseStoreTarget]] = None,
    infer_options: InferOptions = InferOptions.default(),
    mlrun_context=None,
    namespace=None,
    overwrite=None,
    return_df=None,
):
    created_spark_context = False
    try:
        import pyspark.sql

        from mlrun.datastore.spark_utils import check_special_columns_exists

        if spark is None or spark is True:
            # create spark context

            if mlrun_context:
                session_name = f"{mlrun_context.name}-{mlrun_context.uid}"
            else:
                session_name = (
                    f"{featureset.metadata.project}-{featureset.metadata.name}"
                )

            spark = (
                pyspark.sql.SparkSession.builder.appName(session_name)
                .config("spark.driver.memory", "2g")
                .config("spark.sql.session.timeZone", "UTC")
                .getOrCreate()
            )
            created_spark_context = True

        timestamp_key = featureset.spec.timestamp_key
        if isinstance(source, pd.DataFrame):
            df = spark.createDataFrame(source)
        elif isinstance(source, pyspark.sql.DataFrame):
            df = source
        else:
            df = source.to_spark_df(spark, time_field=timestamp_key)
        if featureset.spec.graph and featureset.spec.graph.steps:
            df = run_spark_graph(df, featureset, namespace, spark)

        if isinstance(df, Response) and df.status_code != 0:
            raise mlrun.errors.err_for_status_code(
                df.status_code, df.body.split(": ")[1]
            )

        df.persist()

        _infer_from_static_df(df, featureset, options=infer_options)

        key_columns = list(featureset.spec.entities.keys())
        targets = targets or featureset.spec.targets

        targets_to_ingest = copy.deepcopy(targets)
        featureset.update_targets_for_ingest(targets_to_ingest, overwrite=overwrite)

        for target in targets_to_ingest or []:
            if type(target) is DataTargetBase:
                target = get_target_driver(target, featureset)
            target.set_resource(featureset)
            if featureset.spec.passthrough and target.is_offline:
                check_special_columns_exists(
                    spark_df=df,
                    entities=featureset.spec.entities,
                    timestamp_key=timestamp_key,
                    label_column=featureset.spec.label_column,
                )
                continue
            spark_options = target.get_spark_options(
                key_columns, timestamp_key, overwrite
            )

            df_to_write = df
            df_to_write = target.prepare_spark_df(
                df_to_write, key_columns, timestamp_key, spark_options
            )
            write_format = spark_options.pop("format", None)
            # We can get to this point if the column exists in different letter cases,
            # so PySpark will be able to read it, but we still have to raise an exception for it.

            # This check is here and not in to_spark_df because in spark_merger we can have a target
            # that has different letter cases than the source, like in SnowflakeTarget.
            check_special_columns_exists(
                spark_df=df_to_write,
                entities=featureset.spec.entities,
                timestamp_key=timestamp_key,
                label_column=featureset.spec.label_column,
            )
            if overwrite:
                write_spark_dataframe_with_options(
                    spark_options, df_to_write, "overwrite", write_format=write_format
                )
            else:
                # appending an empty dataframe may cause an empty file to be created (e.g. when writing to parquet)
                # we would like to avoid that
                df_to_write.persist()
                if df_to_write.count() > 0:
                    write_spark_dataframe_with_options(
                        spark_options, df_to_write, "append", write_format=write_format
                    )
            target.update_resource_status("ready")

        if isinstance(source, BaseSourceDriver) and source.schedule:
            max_time = df.agg({timestamp_key: "max"}).collect()[0][0]
            if not max_time:
                # if max_time is None(no data), next scheduled run should be with same start_time
                max_time = source.start_time
            for target in featureset.status.targets:
                featureset.status.update_last_written_for_target(
                    target.get_path().get_absolute_path(
                        project_name=featureset.metadata.project
                    ),
                    max_time,
                )

        _post_ingestion(mlrun_context, featureset, spark)
    finally:
        if created_spark_context:
            spark.stop()
            # We shouldn't return a dataframe that depends on a stopped context
            df = None
    if return_df:
        return df


def _post_ingestion(context, featureset, spark=None):
    featureset.save()
    if context:
        context.logger.info("ingestion task completed, targets:")
        context.logger.info(f"{featureset.status.targets.to_dict()}")
        context.log_result("featureset", featureset.uri)


def _infer_from_static_df(
    df,
    featureset,
    entity_columns=None,
    options: InferOptions = InferOptions.default(),
    sample_size=None,
):
    """infer feature-set schema & stats from static dataframe (without pipeline)"""
    if hasattr(df, "to_dataframe"):
        if hasattr(df, "time_field"):
            time_field = df.time_field or featureset.spec.timestamp_key
        else:
            time_field = featureset.spec.timestamp_key
        if df.is_iterator():
            # todo: describe over multiple chunks
            df = next(df.to_dataframe(time_field=time_field))
        else:
            df = df.to_dataframe(time_field=time_field)
    inferer = get_infer_interface(df)
    if InferOptions.get_common_options(options, InferOptions.schema()):
        featureset.spec.timestamp_key = inferer.infer_schema(
            df,
            featureset.spec.features,
            featureset.spec.entities,
            featureset.spec.timestamp_key,
            entity_columns,
            options=options,
        )
    if InferOptions.get_common_options(options, InferOptions.Stats):
        featureset.status.stats = inferer.get_stats(
            df, options, sample_size=sample_size
        )
    if InferOptions.get_common_options(options, InferOptions.Preview):
        featureset.status.preview = inferer.get_preview(df)
    return df


def set_task_params(
    featureset: FeatureSet,
    source: DataSource = None,
    targets: Optional[list[DataTargetBase]] = None,
    parameters: Optional[dict] = None,
    infer_options: InferOptions = InferOptions.Null,
    overwrite=None,
):
    """convert ingestion parameters to dict, return source + params dict"""
    source = source or featureset.spec.source
    parameters = parameters or {}
    parameters["infer_options"] = infer_options
    parameters["overwrite"] = overwrite
    parameters["featureset"] = featureset.uri
    if source:
        parameters["source"] = source.to_dict()
    if targets:
        parameters["targets"] = [target.to_dict() for target in targets]
    elif not featureset.spec.targets:
        featureset.set_targets()
    featureset.save()
    return source, parameters


def get_feature_set(uri, project=None):
    """get feature set object from the db

    :param uri:  a feature set uri({project}/{name}[:version])
    :param project:  project name if not specified in uri or not using the current/default
    """
    return get_feature_set_by_uri(uri, project)


def get_feature_vector(uri, project=None):
    """get feature vector object from the db

    :param uri:  a feature vector uri({project}/{name}[:version])
    :param project:  project name if not specified in uri or not using the current/default
    """
    return get_feature_vector_by_uri(uri, project, update=False)


def delete_feature_set(name, project="", tag=None, uid=None, force=False):
    """Delete a :py:class:`~mlrun.feature_store.FeatureSet` object from the DB.

    :param name: Name of the object to delete
    :param project: Name of the object's project
    :param tag: Specific object's version tag
    :param uid: Specific object's uid
    :param force: Delete feature set without purging its targets

    If ``tag`` or ``uid`` are specified, then just the version referenced by them will be deleted. Using both
        is not allowed.
        If none are specified, then all instances of the object whose name is ``name`` will be deleted.
    """
    db = mlrun.get_run_db()
    if not force:
        feature_set = db.get_feature_set(name=name, project=project, tag=tag, uid=uid)
        if feature_set.status.targets:
            raise mlrun.errors.MLRunPreconditionFailedError(
                "delete_feature_set requires targets purging. Use either FeatureSet's purge_targets or the force flag."
            )
    return db.delete_feature_set(name=name, project=project, tag=tag, uid=uid)


def delete_feature_vector(name, project="", tag=None, uid=None):
    """Delete a :py:class:`~mlrun.feature_store.FeatureVector` object from the DB.

    :param name: Name of the object to delete
    :param project: Name of the object's project
    :param tag: Specific object's version tag
    :param uid: Specific object's uid

    If ``tag`` or ``uid`` are specified, then just the version referenced by them will be deleted. Using both
        is not allowed.
        If none are specified, then all instances of the object whose name is ``name`` will be deleted.
    """
    db = mlrun.get_run_db()
    return db.delete_feature_vector(name=name, project=project, tag=tag, uid=uid)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import copy
from typing import Optional

import pandas as pd

import mlrun
import mlrun.frameworks

from .artifacts import Artifact, dict_to_artifact
from .config import config
from .render import artifacts_to_html, runs_to_html
from .utils import flatten, get_artifact_target, get_in

list_header = [
    "project",
    "uid",
    "iter",
    "start",
    "end",
    "state",
    "kind",
    "name",
    "labels",
    "inputs",
    "parameters",
    "results",
    "artifacts",
    "artifact_uris",
    "error",
]

iter_index = list_header.index("iter")
state_index = list_header.index("state")
parameters_index = list_header.index("parameters")
results_index = list_header.index("results")


class RunList(list):
    def to_rows(self, extend_iterations=False):
        """return the run list as flattened rows"""
        rows = []
        for run in self:
            iterations = get_in(run, "status.iterations", "")
            row = [
                get_in(run, "metadata.project", config.active_project),
                get_in(run, "metadata.uid", ""),
                get_in(run, "metadata.iteration", ""),
                get_in(run, "status.start_time", ""),
                get_in(run, "status.end_time", ""),
                get_in(run, "status.state", ""),
                get_in(run, "step_kind", get_in(run, "kind", "")),
                get_in(run, "metadata.name", ""),
                get_in(run, "metadata.labels", ""),
                get_in(run, "spec.inputs", ""),
                get_in(run, "spec.parameters", ""),
                get_in(run, "status.results", ""),
                get_in(run, "status.artifacts", []),
                get_in(run, "status.artifact_uris", {}),
                get_in(run, "status.error", ""),
            ]
            if extend_iterations and iterations:
                parameters_dict = {
                    key[len("param.") :]: i
                    for i, key in enumerate(iterations[0])
                    if key.startswith("param.")
                }
                results_dict = {
                    key[len("output.") :]: i
                    for i, key in enumerate(iterations[0])
                    if key.startswith("output.")
                }
                for iter in iterations[1:]:
                    row[state_index] = iter[0]
                    row[iter_index] = iter[1]
                    row[parameters_index] = {
                        key: iter[col] for key, col in parameters_dict.items()
                    }
                    row[results_index] = {
                        key: iter[col] for key, col in results_dict.items()
                    }
                    rows.append(copy(row))
            else:
                rows.append(row)

        return [list_header] + rows

    def to_df(
        self, flat: bool = False, extend_iterations: bool = False, cache: bool = True
    ) -> pd.DataFrame:
        """convert the run list to a dataframe"""
        if hasattr(self, "_df") and cache:
            return self._df
        rows = self.to_rows(extend_iterations=extend_iterations)
        df = pd.DataFrame(rows[1:], columns=rows[0])  # .set_index('iter')
        for time_column in ["start", "end"]:
            df[time_column] = pd.to_datetime(df[time_column])

        if flat:
            df = flatten(df, "labels")
            df = flatten(df, "parameters", "param.")
            df = flatten(df, "results", "output.")
        self._df = df
        return df

    def show(self, display=True, classes=None, short=False, extend_iterations=False):
        """show the run list as a table in Jupyter"""
        html = runs_to_html(
            self.to_df(extend_iterations=extend_iterations),
            display,
            classes=classes,
            short=short,
        )
        if not display:
            return html

    def to_objects(self) -> list["mlrun.RunObject"]:
        """Return a list of Run Objects"""
        return [mlrun.RunObject.from_dict(run) for run in self]

    def compare(
        self,
        hide_identical: bool = True,
        exclude: Optional[list] = None,
        show: Optional[bool] = None,
        extend_iterations=True,
        filename=None,
        colorscale: Optional[str] = None,
    ):
        """return/show parallel coordinates plot + table to compare between the list of runs

        example:

            # return a list of runs in the project, and compare them (show charts)
            runs = project.list_runs(name='download', labels='owner=admin')
            runs.compare()

        :param hide_identical: hide columns with identical values
        :param exclude:        User-provided list of parameters to be excluded from the plot
        :param show:           Allows the user to display the plot within the notebook
        :param extend_iterations: include the iteration (hyper-param) results
        :param filename:       Output filename to save the plot html file
        :param colorscale:     colors used for the lines in the parallel coordinate plot
        :return:  plot html
        """
        return mlrun.frameworks.parallel_coordinates.compare_run_objects(
            self.to_objects(),
            hide_identical=hide_identical,
            exclude=exclude,
            show=show,
            extend_iterations=extend_iterations,
            filename=filename,
            colorscale=colorscale,
        )


class ArtifactList(list):
    def __init__(self, *args):
        super().__init__(*args)
        self.tag = ""

    def to_rows(self):
        """return the artifact list as flattened rows"""
        rows = []
        head = {
            "tree": ["tree", "metadata.tree"],
            "key": ["key", "metadata.key"],
            "iter": ["iter", "metadata.iter"],
            "kind": ["kind", "kind"],
            "path": ["target_path", "spec.target_path"],
            "hash": ["hash", "metadata.hash"],
            "viewer": ["viewer", "spec.viewer"],
            "updated": ["updated", "metadata.updated"],
            "description": ["description", "metadata.description"],
            "producer": ["producer", "spec.producer"],
            "sources": ["sources", "spec.sources"],
            "labels": ["labels", "metadata.labels"],
            # important: the uri item must be the last one in this dict since there is no artifact.uri, and we fill it
            # in the following for loop as the "last_index" in the dict
            "uri": ["uri", "uri"],
        }
        for artifact in self:
            fields_index = 1
            row = [get_in(artifact, v[fields_index], "") for k, v in head.items()]
            artifact_uri = dict_to_artifact(artifact).uri
            last_index = len(row) - 1
            row[last_index] = artifact_uri
            rows.append(row)

        return [head.keys()] + rows

    def to_df(self, flat=False):
        """convert the artifact list to a dataframe"""
        rows = self.to_rows()
        df = pd.DataFrame(rows[1:], columns=rows[0])
        df["updated"] = pd.to_datetime(df["updated"])

        if flat:
            df = flatten(df, "producer", "prod_")
            df = flatten(df, "sources", "src_")

        return df

    def show(self, display=True, classes=None):
        """show the artifact list as a table in Jupyter"""
        df = self.to_df()
        if self.tag != "*":
            df.drop("tree", axis=1, inplace=True)
        html = artifacts_to_html(df, display, classes=classes)
        if not display:
            return html

    def to_objects(self) -> list[Artifact]:
        """return as a list of artifact objects"""
        return [dict_to_artifact(artifact) for artifact in self]

    def dataitems(self) -> list["mlrun.DataItem"]:
        """return as a list of DataItem objects"""
        dataitems = []
        for item in self:
            artifact = get_artifact_target(item)
            if artifact:
                dataitems.append(mlrun.get_dataitem(artifact))
        return dataitems



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from typing import Optional, Union

import pandas as pd
import requests

import mlrun.common.constants as mlrun_constants
import mlrun.common.schemas.schedule
import mlrun.db
import mlrun.errors
import mlrun.launcher.client as launcher
import mlrun.run
import mlrun.runtimes
import mlrun.runtimes.generators
import mlrun.utils.clones
import mlrun.utils.notifications
from mlrun.utils import logger


class ClientRemoteLauncher(launcher.ClientBaseLauncher):
    def launch(
        self,
        runtime: "mlrun.runtimes.KubejobRuntime",
        task: Optional[
            Union["mlrun.run.RunTemplate", "mlrun.run.RunObject", dict]
        ] = None,
        handler: Optional[str] = None,
        name: Optional[str] = "",
        project: Optional[str] = "",
        params: Optional[dict] = None,
        inputs: Optional[dict[str, str]] = None,
        out_path: Optional[str] = "",
        workdir: Optional[str] = "",
        artifact_path: Optional[str] = "",
        output_path: Optional[str] = "",
        watch: Optional[bool] = True,
        schedule: Optional[
            Union[str, mlrun.common.schemas.schedule.ScheduleCronTrigger]
        ] = None,
        hyperparams: Optional[dict[str, list]] = None,
        hyper_param_options: Optional[mlrun.model.HyperParamOptions] = None,
        verbose: Optional[bool] = None,
        scrape_metrics: Optional[bool] = None,
        local_code_path: Optional[str] = None,
        auto_build: Optional[bool] = None,
        param_file_secrets: Optional[dict[str, str]] = None,
        notifications: Optional[list[mlrun.model.Notification]] = None,
        returns: Optional[list[Union[str, dict[str, str]]]] = None,
        state_thresholds: Optional[dict[str, int]] = None,
        reset_on_run: Optional[bool] = None,
        retry: Optional[Union[mlrun.model.Retry, dict]] = None,
    ) -> "mlrun.run.RunObject":
        self.enrich_runtime(runtime, project)
        run = self._create_run_object(task)

        run = self._enrich_run(
            runtime=runtime,
            run=run,
            handler=handler,
            project_name=project,
            name=name,
            params=params,
            inputs=inputs,
            returns=returns,
            hyperparams=hyperparams,
            hyper_param_options=hyper_param_options,
            verbose=verbose,
            scrape_metrics=scrape_metrics,
            output_path=output_path,
            workdir=workdir,
            notifications=notifications,
            state_thresholds=state_thresholds,
            retry=retry,
        )
        self._validate_run(runtime, run)

        if not runtime.is_deployed():
            if runtime.spec.build.auto_build or auto_build:
                logger.info(
                    "Function is not deployed and auto_build flag is set, starting deploy..."
                )
                runtime.deploy(skip_deployed=True, show_on_failure=True)

            else:
                if runtime.requires_build():
                    logger.warning(
                        "Function image is not built/ready and function requires build - execution will fail. "
                        "Need to set auto_build=True or use .deploy() method first"
                    )

        if runtime.verbose:
            logger.info(f"runspec:\n{run.to_yaml()}")

        if (
            "V3IO_USERNAME" in os.environ
            and mlrun_constants.MLRunInternalLabels.v3io_user not in run.metadata.labels
        ):
            run.metadata.labels[mlrun_constants.MLRunInternalLabels.v3io_user] = (
                os.environ.get("V3IO_USERNAME")
            )

        logger.info(
            "Storing function",
            name=run.metadata.name,
            uid=run.metadata.uid,
            db=runtime.spec.rundb,
        )
        self._store_function(runtime, run)

        return self._submit_job(runtime, run, schedule, watch)

    def _submit_job(
        self,
        runtime: "mlrun.runtimes.KubejobRuntime",
        run: "mlrun.run.RunObject",
        schedule: Optional[mlrun.common.schemas.ScheduleCronTrigger] = None,
        watch: Optional[bool] = None,
    ):
        if runtime._secrets:
            run.spec.secret_sources = runtime._secrets.to_serial()
        try:
            db = runtime._get_db()
            resp = db.submit_job(run, schedule=schedule)
            if schedule:
                action = resp.pop("action", "created")
                logger.info(f"Task schedule {action}", **resp)
                return

        except (requests.HTTPError, Exception) as err:
            logger.error("Failed remote run", error=mlrun.errors.err_to_str(err))

            if isinstance(err, requests.HTTPError):
                runtime._handle_submit_job_http_error(err)

            result = None
            # if we got a schedule no reason to do post_run stuff (it purposed to update the run status with error,
            # but there's no run in case of schedule)
            if not schedule:
                result = runtime._update_run_state(
                    task=run, err=mlrun.errors.err_to_str(err)
                )
            return self._wrap_run_result(
                runtime, result, run, schedule=schedule, err=err
            )

        if resp:
            txt = mlrun.runtimes.utils.helpers.get_in(resp, "status.status_text")
            if txt:
                logger.info(txt)

        # watch is None only in scenario where we run from pipeline step, in this case we don't want to watch the run
        # logs too frequently but rather just pull the state of the run from the DB and pull the logs every x seconds
        # which ideally greater than the pull state interval, this reduces unnecessary load on the API server, as
        # running a pipeline is mostly not an interactive process which means the logs pulling doesn't need to be pulled
        # in real time
        if (
            watch is None
            and runtime.kfp
            and mlrun.mlconf.httpdb.logs.pipelines.pull_state.mode == "enabled"
        ):
            state_interval = int(
                mlrun.mlconf.httpdb.logs.pipelines.pull_state.pull_state_interval
            )
            logs_interval = int(
                mlrun.mlconf.httpdb.logs.pipelines.pull_state.pull_logs_interval
            )
            run.wait_for_completion(
                show_logs=True,
                sleep=state_interval,
                logs_interval=logs_interval,
                raise_on_failure=False,
            )
            resp = runtime._get_db_run(run)

        elif watch or runtime.kfp:
            run.logs(True, runtime._get_db())
            resp = runtime._get_db_run(run)

        return self._wrap_run_result(runtime, resp, run, schedule=schedule)

    @classmethod
    def _validate_run_single_param(cls, param_name: str, param_value: int):
        if isinstance(param_value, pd.DataFrame):
            raise mlrun.errors.MLRunInvalidArgumentTypeError(
                f"Parameter '{param_name}' has an unsupported value of type"
                f" 'pandas.DataFrame' in remote execution."
            )
        super()._validate_run_single_param(
            param_name=param_name, param_value=param_value
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import datetime
import typing

import mlrun
import mlrun.common.model_monitoring.helpers
import mlrun.feature_store as fstore
import mlrun.feature_store.steps
import mlrun.serving.states
import mlrun.utils
from mlrun.common.schemas.model_monitoring.constants import (
    ControllerEvent,
    ControllerEventKind,
    EndpointType,
    EventFieldType,
    FileTargetKind,
    ProjectSecretKeys,
)
from mlrun.model_monitoring.db import TSDBConnector
from mlrun.utils import logger


# Stream processing code
class EventStreamProcessor:
    def __init__(
        self,
        project: str,
        parquet_batching_max_events: int,
        parquet_batching_timeout_secs: int,
        parquet_target: str,
        aggregate_windows: typing.Optional[list[str]] = None,
        aggregate_period: str = "5m",
        model_monitoring_access_key: typing.Optional[str] = None,
    ):
        # General configurations, mainly used for the storey steps in the future serving graph
        self.project = project
        self.aggregate_windows = aggregate_windows or ["5m", "1h"]
        self.aggregate_period = aggregate_period

        # Parquet path and configurations
        self.parquet_path = parquet_target
        self.parquet_batching_max_events = parquet_batching_max_events
        self.parquet_batching_timeout_secs = parquet_batching_timeout_secs

        logger.info(
            "Initializing model monitoring event stream processor",
            parquet_path=self.parquet_path,
            parquet_batching_max_events=self.parquet_batching_max_events,
        )

        self.tsdb_configurations = {}
        if not mlrun.mlconf.is_ce_mode():
            self._initialize_v3io_configurations(
                model_monitoring_access_key=model_monitoring_access_key
            )

    def _initialize_v3io_configurations(
        self,
        tsdb_batching_max_events: int = 10,
        tsdb_batching_timeout_secs: int = 60 * 5,  # Default 5 minutes
        v3io_access_key: typing.Optional[str] = None,
        v3io_framesd: typing.Optional[str] = None,
        v3io_api: typing.Optional[str] = None,
        model_monitoring_access_key: typing.Optional[str] = None,
    ):
        # Get the V3IO configurations
        self.v3io_framesd = v3io_framesd or mlrun.mlconf.v3io_framesd
        self.v3io_api = v3io_api or mlrun.mlconf.v3io_api

        self.v3io_access_key = v3io_access_key or mlrun.mlconf.get_v3io_access_key()
        self.model_monitoring_access_key = (
            model_monitoring_access_key
            or mlrun.get_secret_or_env(ProjectSecretKeys.ACCESS_KEY)
            or self.v3io_access_key
        )

        # TSDB path and configurations
        tsdb_path = mlrun.mlconf.get_model_monitoring_file_target_path(
            project=self.project, kind=FileTargetKind.EVENTS
        )
        (
            _,
            self.tsdb_container,
            self.tsdb_path,
        ) = mlrun.common.model_monitoring.helpers.parse_model_endpoint_store_prefix(
            tsdb_path
        )

        self.tsdb_path = f"{self.tsdb_container}/{self.tsdb_path}"
        self.tsdb_batching_max_events = tsdb_batching_max_events
        self.tsdb_batching_timeout_secs = tsdb_batching_timeout_secs

    def apply_monitoring_serving_graph(
        self,
        fn: mlrun.runtimes.ServingRuntime,
        tsdb_connector: TSDBConnector,
        controller_stream_uri: str,
    ) -> None:
        """
        Apply monitoring serving graph to a given serving function. The following serving graph includes about 4 main
        parts that each one them includes several steps of different operations that are executed on the events from
        the model server.
        Each event has metadata (function_uri, timestamp, class, etc.) but also inputs, predictions and optional
        metrics from the model server.
        In ths first part, the serving graph processes the event and splits it into sub-events. This part also includes
        validation of the event data and adding important details to the event such as endpoint_id.
        In the next parts, the serving graph stores data to 3 different targets:
        1. KV/SQL: Metadata and basic stats about the average latency and the amount of predictions over
           time per endpoint. for example the amount of predictions of endpoint x in the last 5 min. The model
           endpoints table also contains data on the model endpoint from other processes, such as feature_stats that
           represents sample statistics from the training data. If the target is from type KV, then the model endpoints
           table can be found under v3io:///users/pipelines/project-name/model-endpoints/endpoints/. If the target is
           SQL, then the table is stored within the database that was defined in the provided connection string.
        2. TSDB: live data of different key metric dictionaries in tsdb target.
           This data is being used by the monitoring dashboards in grafana. If using V3IO TSDB, results
           can be found under  v3io:///users/pipelines/project-name/model-endpoints/events/. In that case, we generate
           3 different key  metric dictionaries: base_metrics (average latency and predictions over time),
           endpoint_features (Prediction and feature names and values), and custom_metrics (user-defined metrics).
        3. Parquet: This Parquet file includes the required data for the model monitoring applications. If defined,
           the parquet target path can be found under mlrun.mlconf.model_endpoint_monitoring.offline. Otherwise,
           the default parquet path is under mlrun.mlconf.model_endpoint_monitoring.user_space. Note that if you are
           using CE, the parquet target path is based on the defined MLRun artifact path.

        In a separate branch, "batch complete" events are forwarded to the controller stream with an intentional delay,
        to allow for data to first be written to parquet.

        :param fn: A serving function.
        :param tsdb_connector: Time series database connector.
        :param controller_stream_uri: The controller stream URI. Runs on server api pod so needed to be provided as
        input
        """

        graph = typing.cast(
            mlrun.serving.states.RootFlowStep,
            fn.set_topology(mlrun.serving.states.StepKinds.flow, engine="async"),
        )

        # forward back complete events to controller
        graph.add_step(
            "storey.Filter",
            "FilterBatchComplete",
            _fn="(event.get('kind') == 'batch_complete')",
        )

        graph.add_step(
            "Delay",
            name="BatchDelay",
            after="FilterBatchComplete",
            delay=self.parquet_batching_timeout_secs + 5,  # add margin
        )

        # split the graph between event with error vs valid event
        graph.add_step(
            "storey.Filter",
            "FilterError",
            _fn="(event.get('error') is None)",
        )

        graph.add_step(
            "storey.Filter",
            "ForwardError",
            _fn="(event.get('error') is not None)",
        )

        tsdb_connector.handle_model_error(
            graph,
        )

        # Process endpoint event: splitting into sub-events and validate event data
        def apply_process_endpoint_event():
            graph.add_step(
                "ProcessEndpointEvent",
                after="FilterError",
                full_event=True,
                project=self.project,
            )

        apply_process_endpoint_event()

        # Applying Storey operations of filtering and flatten
        def apply_storey_filter_and_flatmap():
            # Remove none values from each event
            graph.add_step(
                "storey.Filter",
                "filter_none",
                _fn="(event is not None)",
                after="ProcessEndpointEvent",
            )

            # flatten the events
            graph.add_step(
                "storey.FlatMap", "flatten_events", _fn="(event)", after="filter_none"
            )

        apply_storey_filter_and_flatmap()

        # Validating feature names and map each feature to its value
        def apply_map_feature_names():
            graph.add_step(
                "MapFeatureNames",
                name="MapFeatureNames",
                infer_columns_from_data=True,
                project=self.project,
                after="flatten_events",
            )

        apply_map_feature_names()
        # split the graph between event with error vs valid event
        graph.add_step(
            "storey.Filter",
            "FilterNOP",
            after="MapFeatureNames",
            _fn="(event.get('kind', " ") != 'nop_event')",
        )
        graph.add_step(
            "storey.Filter",
            "ForwardNOP",
            after="MapFeatureNames",
            _fn="(event.get('kind', " ") == 'nop_event')",
        )

        tsdb_connector.apply_monitoring_stream_steps(
            graph=graph,
            aggregate_windows=self.aggregate_windows,
            aggregate_period=self.aggregate_period,
        )

        # Parquet branch
        # Filter and validate different keys before writing the data to Parquet target
        def apply_process_before_parquet():
            graph.add_step(
                "ProcessBeforeParquet",
                name="ProcessBeforeParquet",
                after="FilterNOP",
                _fn="(event)",
            )

        apply_process_before_parquet()

        # Write the Parquet target file, partitioned by key (endpoint_id) and time.
        def apply_parquet_target():
            graph.add_step(
                "mlrun.datastore.storeytargets.ParquetStoreyTarget",
                alternative_v3io_access_key=mlrun.common.schemas.model_monitoring.ProjectSecretKeys.ACCESS_KEY,
                name="ParquetTarget",
                after="ProcessBeforeParquet",
                graph_shape="cylinder",
                path=self.parquet_path,
                max_events=self.parquet_batching_max_events,
                flush_after_seconds=self.parquet_batching_timeout_secs,
                attributes={"infer_columns_from_data": True},
                index_cols=[EventFieldType.ENDPOINT_ID],
                key_bucketing_number=0,
                time_partitioning_granularity="hour",
                time_field=EventFieldType.TIMESTAMP,
                partition_cols=["$key", "$year", "$month", "$day", "$hour"],
            )

        apply_parquet_target()

        # controller branch
        def apply_push_controller_stream(stream_uri: str):
            graph.add_step(
                ">>",
                "controller_stream",
                path=stream_uri,
                sharding_func=ControllerEvent.ENDPOINT_ID,
                after=["ForwardNOP", "BatchDelay"],
                # Force using the pipeline key instead of the one in the profile in case of v3io profile.
                # In case of Kafka, this parameter will be ignored.
                alternative_v3io_access_key="V3IO_ACCESS_KEY",
            )

        apply_push_controller_stream(controller_stream_uri)


class ProcessBeforeParquet(mlrun.feature_store.steps.MapClass):
    def __init__(self, **kwargs):
        """
        Process the data before writing to Parquet file. In this step, unnecessary keys will be removed while possible
        missing keys values will be set to None.

        :returns: Event dictionary with filtered data for the Parquet target.

        """
        super().__init__(**kwargs)

    def do(self, event):
        logger.info("ProcessBeforeParquet1", event=event)
        # Remove the following keys from the event
        for key in [
            EventFieldType.FEATURES,
            EventFieldType.NAMED_FEATURES,
            EventFieldType.PREDICTION,
            EventFieldType.NAMED_PREDICTIONS,
        ]:
            event.pop(key, None)

        # Split entities dictionary to separate dictionaries within the event
        value = event.get("entities")
        if value is not None:
            event = {**value, **event}

        # Validate that the following keys exist
        for key in [
            EventFieldType.LABELS,
            EventFieldType.METRICS,
            EventFieldType.ENTITIES,
        ]:
            if not event.get(key):
                event[key] = None
        logger.info("ProcessBeforeParquet2", event=event)
        return event


class Delay(mlrun.feature_store.steps.MapClass):
    def __init__(self, delay: int, **kwargs):
        super().__init__(**kwargs)
        self._delay = delay

    async def do(self, event):
        await asyncio.sleep(self._delay)
        return event


class ProcessEndpointEvent(mlrun.feature_store.steps.MapClass):
    def __init__(
        self,
        project: str,
        **kwargs,
    ):
        """
        Process event or batch of events as part of the first step of the monitoring serving graph. It includes
        Adding important details to the event such as endpoint_id, handling errors coming from the stream, validation
        of event data such as inputs and outputs, and splitting model event into sub-events.

        :param project: Project name.

        :returns: A Storey event object which is the basic unit of data in Storey. Note that the next steps of
                  the monitoring serving graph are based on Storey operations.

        """
        super().__init__(**kwargs)

        self.project: str = project

        # First and last requests timestamps (value) of each endpoint (key)
        self.first_request: dict[str, str] = dict()
        self.last_request: dict[str, str] = dict()

        # Set of endpoints in the current events
        self.endpoints: set[str] = set()

    def do(self, full_event):
        event = full_event.body
        if event.get(ControllerEvent.KIND, "") == ControllerEventKind.NOP_EVENT:
            logger.debug(
                "Skipped nop event inside of ProcessEndpointEvent", event=event
            )
            full_event.body = [event]
            return full_event
        # Getting model version and function uri from event
        # and use them for retrieving the endpoint_id
        function_uri = full_event.body.get(EventFieldType.FUNCTION_URI)
        if not is_not_none(function_uri, [EventFieldType.FUNCTION_URI]):
            return None

        model = full_event.body.get(EventFieldType.MODEL)
        if not is_not_none(model, [EventFieldType.MODEL]):
            return None

        endpoint_id = event[EventFieldType.ENDPOINT_ID]

        # In case this process fails, resume state from existing record
        self.resume_state(
            endpoint_id=endpoint_id,
            endpoint_name=full_event.body.get(EventFieldType.MODEL),
        )

        # Validate event fields
        model_class = event.get("model_class") or event.get("class")
        timestamp = event.get("when")
        request_id = event.get("request", {}).get("id") or event.get("resp", {}).get(
            "id"
        )
        feature_names = event.get("request", {}).get("input_schema")
        labels_names = event.get("resp", {}).get("output_schema")
        latency = event.get("microsec")
        features = event.get("request", {}).get("inputs")
        predictions = event.get("resp", {}).get("outputs")

        if not self.is_valid(
            validation_function=is_not_none,
            field=timestamp,
            dict_path=["when"],
        ):
            return None

        if endpoint_id not in self.first_request:
            # Set time for the first request of the current endpoint
            self.first_request[endpoint_id] = timestamp

        if not self.is_valid(
            validation_function=is_not_none,
            field=request_id,
            dict_path=["request", "id"],
        ):
            return None
        if not self.is_valid(
            validation_function=is_not_none,
            field=latency,
            dict_path=["microsec"],
        ):
            return None
        if not self.is_valid(
            validation_function=is_not_none,
            field=features,
            dict_path=["request", "inputs"],
        ):
            return None
        if not self.is_valid(
            validation_function=is_not_none,
            field=predictions,
            dict_path=["resp", "outputs"],
        ):
            return None

        # Convert timestamp to a datetime object
        timestamp_obj = datetime.datetime.fromisoformat(timestamp)

        # Separate each model invocation into sub events that will be stored as dictionary
        # in list of events. This list will be used as the body for the storey event.
        if not isinstance(features, list):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Model's inputs must be a list"
            )
        features = (
            features
            if not any(not isinstance(feat, list) for feat in features)
            else [features]
        )
        if not isinstance(predictions, list):
            predictions = [[predictions]]
        elif isinstance(predictions, list) and len(predictions) == len(features):
            pass  # predictions are already in the right format
        else:
            predictions = (
                predictions
                if not any(not isinstance(pred, list) for pred in predictions)
                else [predictions]
            )

        events = []
        for i, (feature, prediction) in enumerate(zip(features, predictions)):
            if not isinstance(prediction, list):
                prediction = [prediction]

            if not isinstance(feature, list):
                feature = [feature]

            effective_sample_count, estimated_prediction_count = (
                self._get_effective_and_estimated_counts(event=event)
            )

            events.append(
                {
                    EventFieldType.FUNCTION_URI: function_uri,
                    EventFieldType.ENDPOINT_NAME: event.get(EventFieldType.MODEL),
                    EventFieldType.MODEL_CLASS: model_class,
                    EventFieldType.TIMESTAMP: timestamp_obj,
                    EventFieldType.ENDPOINT_ID: endpoint_id,
                    EventFieldType.REQUEST_ID: request_id,
                    EventFieldType.LATENCY: latency,
                    EventFieldType.FEATURES: feature,
                    EventFieldType.PREDICTION: prediction,
                    EventFieldType.FIRST_REQUEST: self.first_request[endpoint_id],
                    EventFieldType.LAST_REQUEST: timestamp,
                    EventFieldType.LAST_REQUEST_TIMESTAMP: mlrun.utils.enrich_datetime_with_tz_info(
                        timestamp
                    ).timestamp(),
                    EventFieldType.LABELS: event.get(EventFieldType.LABELS, {}),
                    EventFieldType.METRICS: event.get(EventFieldType.METRICS, {}),
                    EventFieldType.ENTITIES: event.get("request", {}).get(
                        EventFieldType.ENTITIES, {}
                    ),
                    EventFieldType.EFFECTIVE_SAMPLE_COUNT: effective_sample_count,
                    EventFieldType.ESTIMATED_PREDICTION_COUNT: estimated_prediction_count,
                    EventFieldType.FEATURE_NAMES: feature_names,
                    EventFieldType.LABEL_NAMES: labels_names,
                }
            )

        # Create a storey event object with list of events, based on endpoint_id which will be used
        # in the upcoming steps
        full_event.key = endpoint_id
        full_event.body = events
        return full_event

    def resume_state(self, endpoint_id, endpoint_name):
        # Make sure process is resumable, if process fails for any reason, be able to pick things up close to where we
        # left them
        if endpoint_id not in self.endpoints:
            logger.info("Trying to resume state", endpoint_id=endpoint_id)
            endpoint_record = (
                mlrun.db.get_run_db()
                .get_model_endpoint(
                    project=self.project,
                    endpoint_id=endpoint_id,
                    name=endpoint_name,
                    tsdb_metrics=False,
                )
                .flat_dict()
            )

            # If model endpoint found, get first_request & last_request values
            if endpoint_record:
                first_request = endpoint_record.get(EventFieldType.FIRST_REQUEST)

                if first_request:
                    self.first_request[endpoint_id] = first_request

            # add endpoint to endpoints set
            self.endpoints.add(endpoint_id)

    def is_valid(
        self,
        validation_function,
        field: typing.Any,
        dict_path: list[str],
    ):
        if validation_function(field, dict_path):
            return True

        return False

    @staticmethod
    def _get_effective_and_estimated_counts(event):
        """
        Calculate the `effective_sample_count` and the `estimated_prediction_count` based on the event's
        sampling percentage. These values will be stored in the TSDB target.
        Note that In non-batch serving, the `effective_sample_count` is always set to 1. In addition, when the sampling
        percentage is 100%, the `estimated_prediction_count` is equal to the `effective_sample_count`.
        """
        effective_sample_count = event.get(EventFieldType.EFFECTIVE_SAMPLE_COUNT, 1)
        estimated_prediction_count = effective_sample_count * (
            100 / event.get(EventFieldType.SAMPLING_PERCENTAGE, 100)
        )
        return effective_sample_count, estimated_prediction_count


def is_not_none(field: typing.Any, dict_path: list[str]):
    if field is not None:
        return True
    logger.error(
        f"Expected event field is missing: {field} [Event -> {','.join(dict_path)}]"
    )
    return False


class MapFeatureNames(mlrun.feature_store.steps.MapClass):
    def __init__(
        self,
        project: str,
        infer_columns_from_data: bool = False,
        **kwargs,
    ):
        """
        Validating feature names and label columns and map each feature to its value. In the end of this step,
        the event should have key-value pairs of (feature name: feature value).

        :param project:                 Project name.
        :param infer_columns_from_data: If true and features or labels names were not found, then try to
                                        retrieve them from data that was stored in the previous events of
                                        the current process. This data can be found under self.feature_names and
                                        self.label_columns.


        :returns: A single event as a dictionary that includes metadata (endpoint_id, model_class, etc.) and also
                  feature names and values (as well as the prediction results).
        """
        super().__init__(**kwargs)

        self._infer_columns_from_data = infer_columns_from_data
        self.project = project

        # Dictionaries that will be used in case features names
        # and labels columns were not found in the current event
        self.feature_names = {}
        self.label_columns = {}
        self.first_request = {}

        # Dictionary to manage the model endpoint types - important for the V3IO TSDB
        self.endpoint_type = {}

    def _infer_feature_names_from_data(self, event):
        endpoint_id = event[EventFieldType.ENDPOINT_ID]
        if endpoint_id in self.feature_names and len(
            self.feature_names[endpoint_id]
        ) >= len(event[EventFieldType.FEATURES]):
            return self.feature_names[endpoint_id]
        return None

    def _infer_label_columns_from_data(self, event):
        endpoint_id = event[EventFieldType.ENDPOINT_ID]
        if endpoint_id in self.label_columns and len(
            self.label_columns[endpoint_id]
        ) >= len(event[EventFieldType.PREDICTION]):
            return self.label_columns[endpoint_id]
        return None

    def do(self, event: dict):
        if event.get(ControllerEvent.KIND, "") == ControllerEventKind.NOP_EVENT:
            return event
        endpoint_id = event[EventFieldType.ENDPOINT_ID]

        feature_values = event[EventFieldType.FEATURES]
        label_values = event[EventFieldType.PREDICTION]

        for index in range(len(feature_values)):
            feature_value = feature_values[index]
            if isinstance(feature_value, int):
                feature_values[index] = float(feature_value)

        attributes_to_update = {}
        endpoint_record = None
        # Get feature names and label columns
        if endpoint_id not in self.feature_names:
            endpoint_record = (
                mlrun.db.get_run_db()
                .get_model_endpoint(
                    project=self.project,
                    endpoint_id=endpoint_id,
                    name=event[EventFieldType.ENDPOINT_NAME],
                    tsdb_metrics=False,
                )
                .flat_dict()
            )
            feature_names = endpoint_record.get(EventFieldType.FEATURE_NAMES)

            label_columns = endpoint_record.get(EventFieldType.LABEL_NAMES)

            # If feature names were not found,
            # try to retrieve them from the previous events of the current process
            if not feature_names and self._infer_columns_from_data:
                feature_names = self._infer_feature_names_from_data(event)

            endpoint_type = int(endpoint_record.get(EventFieldType.ENDPOINT_TYPE))
            if not feature_names:
                logger.warn(
                    "Feature names are not initialized, they will be automatically generated",
                    endpoint_id=endpoint_id,
                )
                feature_names = event.get(EventFieldType.FEATURE_NAMES) or [
                    f"f{i}" for i, _ in enumerate(event[EventFieldType.FEATURES])
                ]

                # Update the endpoint record with the generated features
                attributes_to_update[EventFieldType.FEATURE_NAMES] = feature_names

                if endpoint_type != EndpointType.ROUTER.value:
                    update_monitoring_feature_set(
                        endpoint_record=endpoint_record,
                        feature_names=feature_names,
                        feature_values=feature_values,
                    )

            # Similar process with label columns
            if not label_columns and self._infer_columns_from_data:
                label_columns = self._infer_label_columns_from_data(event)

            if not label_columns:
                logger.warn(
                    "label column names are not initialized, they will be automatically generated",
                    endpoint_id=endpoint_id,
                )
                label_columns = event.get(EventFieldType.LABEL_NAMES) or [
                    f"p{i}" for i, _ in enumerate(event[EventFieldType.PREDICTION])
                ]
                attributes_to_update[EventFieldType.LABEL_NAMES] = label_columns
                if endpoint_type != EndpointType.ROUTER.value:
                    update_monitoring_feature_set(
                        endpoint_record=endpoint_record,
                        feature_names=label_columns,
                        feature_values=label_values,
                    )

            self.label_columns[endpoint_id] = label_columns
            self.feature_names[endpoint_id] = feature_names

            logger.info(
                "Label columns", endpoint_id=endpoint_id, label_columns=label_columns
            )
            logger.info(
                "Feature names", endpoint_id=endpoint_id, feature_names=feature_names
            )

            # Update the endpoint type within the endpoint types dictionary
            self.endpoint_type[endpoint_id] = endpoint_type

        # Update the first request time in the endpoint record
        if endpoint_id not in self.first_request:
            endpoint_record = endpoint_record or (
                mlrun.db.get_run_db()
                .get_model_endpoint(
                    project=self.project,
                    endpoint_id=endpoint_id,
                    name=event[EventFieldType.ENDPOINT_NAME],
                    tsdb_metrics=False,
                )
                .flat_dict()
            )
            if not endpoint_record.get(EventFieldType.FIRST_REQUEST):
                attributes_to_update[EventFieldType.FIRST_REQUEST] = (
                    mlrun.utils.enrich_datetime_with_tz_info(
                        event[EventFieldType.FIRST_REQUEST]
                    )
                )
            self.first_request[endpoint_id] = True

        if attributes_to_update:
            logger.info(
                "Updating endpoint record",
                endpoint_id=endpoint_id,
                attributes=attributes_to_update,
            )
            update_endpoint_record(
                project=self.project,
                endpoint_id=endpoint_id,
                attributes=attributes_to_update,
                endpoint_name=event[EventFieldType.ENDPOINT_NAME],
            )

        # Add feature_name:value pairs along with a mapping dictionary of all of these pairs
        feature_names = self.feature_names[endpoint_id]
        self._map_dictionary_values(
            event=event,
            named_iters=feature_names,
            values_iters=feature_values,
            mapping_dictionary=EventFieldType.NAMED_FEATURES,
        )

        # Add label_name:value pairs along with a mapping dictionary of all of these pairs
        label_names = self.label_columns[endpoint_id]
        self._map_dictionary_values(
            event=event,
            named_iters=label_names,
            values_iters=label_values,
            mapping_dictionary=EventFieldType.NAMED_PREDICTIONS,
        )

        # Add endpoint type to the event
        event[EventFieldType.ENDPOINT_TYPE] = self.endpoint_type[endpoint_id]

        logger.info("Mapped event", event=event)
        return event

    @staticmethod
    def _map_dictionary_values(
        event: dict,
        named_iters: list,
        values_iters: list,
        mapping_dictionary: str,
    ):
        """Adding name-value pairs to event dictionary based on two provided lists of names and values. These pairs
        will be used mainly for the Parquet target file. In addition, this function creates a new mapping dictionary of
        these pairs which will be unpacked in ProcessBeforeTSDB step

        :param event:               A dictionary that includes details about the current event such as endpoint_id
                                    and input names and values.
        :param named_iters:         List of names to match to the list of values.
        :param values_iters:        List of values to match to the list of names.
        :param mapping_dictionary:  Name of the new dictionary that will be stored in the current event. The new
                                    dictionary includes name-value pairs based on the provided named_iters and
                                    values_iters lists.

        """
        event[mapping_dictionary] = {}
        diff = len(named_iters) - len(values_iters)
        values_iters += [None] * diff
        for name, value in zip(named_iters, values_iters):
            event[name] = value
            event[mapping_dictionary][name] = value


class InferSchema(mlrun.feature_store.steps.MapClass):
    def __init__(
        self,
        v3io_framesd: str,
        container: str,
        table: str,
        **kwargs,
    ):
        """
        Apply infer_schema on the kv table which generates the schema file.
        Grafana monitoring dashboards use this schema to query the relevant stats.

        :param v3io_access_key:         Access key with permission to a KV table.
        :v3io_framesd:                  path to v3io frames.
        :param container:               Name of the container that will be used to retrieve the endpoint id. For model
                                        endpoints it is usually 'users'.
        :param table:                   KV table path that will be used to retrieve the endpoint id. For model endpoints
                                        it is usually pipelines/project-name/model-endpoints/endpoints/.

        """
        super().__init__(**kwargs)
        self.container = container
        self.v3io_framesd = v3io_framesd
        self.table = table
        self.keys = set()

    def do(self, event: dict):
        key_set = set(event.keys())
        if not key_set.issubset(self.keys):
            import mlrun.utils.v3io_clients

            self.keys.update(key_set)
            # Apply infer_schema on the kv table for generating the schema file
            mlrun.utils.v3io_clients.get_frames_client(
                container=self.container,
                address=self.v3io_framesd,
            ).execute(backend="kv", table=self.table, command="infer_schema")

        return event


def update_endpoint_record(
    project: str,
    endpoint_id: str,
    endpoint_name: str,
    attributes: dict,
):
    mlrun.db.get_run_db().patch_model_endpoint(
        project=project,
        endpoint_id=endpoint_id,
        attributes=attributes,
        name=endpoint_name,
    )


def update_monitoring_feature_set(
    endpoint_record: dict[str, typing.Any],
    feature_names: list[str],
    feature_values: list[typing.Any],
):
    monitoring_feature_set = fstore.get_feature_set(
        endpoint_record[
            mlrun.common.schemas.model_monitoring.EventFieldType.FEATURE_SET_URI
        ]
    )
    for name, val in zip(feature_names, feature_values):
        monitoring_feature_set.add_feature(
            fstore.Feature(name=name, value_type=type(val))
        )

    monitoring_feature_set.save()



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import typing

import mlrun.common.schemas
import mlrun.db
import mlrun.errors
from mlrun_pipelines.common.ops import build_op

from ..model import RunObject
from .pod import KubeResource


class KubejobRuntime(KubeResource):
    kind = "job"
    _is_nested = True

    _is_remote = True

    def is_deployed(self):
        """check if the function is deployed (has a valid container)"""
        if self.spec.image:
            return True

        db = self._get_db()
        try:
            # getting builder status enriches the runtime when it needs to be fetched from the API,
            # otherwise it's a no-op
            db.get_builder_status(self, logs=False)
        except Exception:
            pass

        if self.spec.image:
            return True
        if self.status.state and self.status.state == "ready":
            return True
        return False

    def with_source_archive(
        self, source, workdir=None, handler=None, pull_at_runtime=True, target_dir=None
    ):
        """load the code from git/tar/zip archive at runtime or build

        :param source:          valid absolute path or URL to git, zip, or tar file, e.g.
                                git://github.com/mlrun/something.git
                                http://some/url/file.zip
                                note path source must exist on the image or exist locally when run is local
                                (it is recommended to use 'workdir' when source is a filepath instead)
        :param handler:         default function handler
        :param workdir:         working dir relative to the archive root (e.g. './subdir') or absolute to the image root
        :param pull_at_runtime: load the archive into the container at job runtime vs on build/deploy
        :param target_dir:      target dir on runtime pod or repo clone / archive extraction
        """
        self._configure_mlrun_build_with_source(
            source=source,
            workdir=workdir,
            handler=handler,
            pull_at_runtime=pull_at_runtime,
            target_dir=target_dir,
        )

    def build_config(
        self,
        image="",
        base_image=None,
        commands: typing.Optional[list] = None,
        secret=None,
        source=None,
        extra=None,
        load_source_on_run=None,
        with_mlrun=None,
        auto_build=None,
        requirements=None,
        overwrite=True,
        prepare_image_for_deploy=True,
        requirements_file=None,
        builder_env=None,
        extra_args=None,
    ):
        """specify builder configuration for the deploy operation

        :param image:      target image name/path
        :param base_image: base image name/path
        :param commands:   list of docker build (RUN) commands e.g. ['pip install pandas']
        :param secret:     k8s secret for accessing the docker registry
        :param source:     source git/tar archive to load code from in to the context/workdir
                           e.g. git://github.com/mlrun/something.git#development
        :param extra:      extra Dockerfile lines
        :param load_source_on_run: load the archive code into the container at runtime vs at build time
        :param with_mlrun: add the current mlrun package to the container build
        :param auto_build: when set to True and the function require build it will be built on the first
                           function run, use only if you dont plan on changing the build config between runs
        :param requirements: a list of packages to install
        :param requirements_file: requirements file to install
        :param overwrite:  overwrite existing build configuration (currently applies to requirements and commands)
           * False: the new params are merged with the existing
           * True: the existing params are replaced by the new ones
        :param prepare_image_for_deploy:    prepare the image/base_image spec for deployment
        :param extra_args:  A string containing additional builder arguments in the format of command-line options,
            e.g. extra_args="--skip-tls-verify --build-arg A=val"
        :param builder_env: Kaniko builder pod env vars dict (for config/credentials)
            e.g. builder_env={"GIT_TOKEN": token}
        """
        image = mlrun.utils.helpers.remove_image_protocol_prefix(image)
        self.spec.build.build_config(
            image=image,
            base_image=base_image,
            commands=commands,
            secret=secret,
            source=source,
            extra=extra,
            load_source_on_run=load_source_on_run,
            with_mlrun=with_mlrun,
            auto_build=auto_build,
            requirements=requirements,
            requirements_file=requirements_file,
            overwrite=overwrite,
            builder_env=builder_env,
            extra_args=extra_args,
        )

        if prepare_image_for_deploy:
            self.prepare_image_for_deploy()

    def deploy(
        self,
        watch: bool = True,
        with_mlrun: typing.Optional[bool] = None,
        skip_deployed: bool = False,
        is_kfp: bool = False,
        mlrun_version_specifier: typing.Optional[bool] = None,
        builder_env: typing.Optional[dict] = None,
        show_on_failure: bool = False,
        force_build: bool = False,
    ) -> bool:
        """Deploy function, build container with dependencies

        :param watch:                   Wait for the deploy to complete (and print build logs)
        :param with_mlrun:              Add the current mlrun package to the container build
        :param skip_deployed:           Skip the build if we already have an image for the function
        :param is_kfp:                  Deploy as part of a kfp pipeline
        :param mlrun_version_specifier: Which mlrun package version to include (if not current)
        :param builder_env:             Kaniko builder pod env vars dict (for config/credentials)
                                        e.g. builder_env={"GIT_TOKEN": token}
        :param show_on_failure:         Show logs only in case of build failure
        :param force_build:             Set True for force building the image, even when no changes were made

        :return: True if the function is ready (deployed)
        """

        build = self.spec.build
        with_mlrun = self._resolve_build_with_mlrun(with_mlrun)

        self.status.state = ""
        if build.base_image:
            # clear the image so build will not be skipped
            self.spec.image = ""

        return self._build_image(
            builder_env=builder_env,
            force_build=force_build,
            mlrun_version_specifier=mlrun_version_specifier,
            show_on_failure=show_on_failure,
            skip_deployed=skip_deployed,
            watch=watch,
            is_kfp=is_kfp,
            with_mlrun=with_mlrun,
        )

    def deploy_step(
        self,
        image=None,
        base_image=None,
        commands: typing.Optional[list] = None,
        secret_name="",
        with_mlrun=True,
        skip_deployed=False,
    ):
        function_name = self.metadata.name or "function"
        name = f"deploy_{function_name}"
        # mark that the function/image is built as part of the pipeline so other places
        # which use the function will grab the updated image/status
        self._build_in_pipeline = True
        return build_op(
            name,
            self,
            image=image,
            base_image=base_image,
            commands=commands,
            secret_name=secret_name,
            with_mlrun=with_mlrun,
            skip_deployed=skip_deployed,
        )

    def _run(self, runobj: RunObject, execution):
        raise NotImplementedError(
            f"Running a {self.kind} function from the client is not supported. Use .run() to submit the job to the API."
        )

    @property
    def serving_spec(self):
        return self.spec.serving_spec



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import inspect
import re

from mlrun.model import FunctionEntrypoint


def type_name(ann, empty_is_none=False):
    if ann is inspect.Signature.empty:
        return None if empty_is_none else ""
    return getattr(ann, "__name__", str(ann))


def inspect_default(value):
    if value is inspect.Signature.empty:
        return None
    return repr(value)


def inspect_param(param: inspect.Parameter) -> dict:
    name = param.name
    typ = type_name(param.annotation)
    default = inspect_default(param.default)
    return param_dict(name, typ, "", default)


# We're using dict and not classes (here and in func_dict) since this goes
# directly to YAML
def param_dict(name="", type="", doc="", default=""):
    return_value = {
        "doc": doc,
        "name": name,
        "type": type,
    }
    if default is not None:
        return_value["default"] = default
    return return_value


def func_dict(
    name,
    doc,
    params,
    returns,
    lineno,
    has_varargs: bool = False,
    has_kwargs: bool = False,
):
    return {
        "name": name,
        "doc": doc,
        "params": params,
        "return": returns,
        "lineno": lineno,
        "has_varargs": has_varargs,
        "has_kwargs": has_kwargs,
    }


def func_lineno(fn):
    try:
        return inspect.getsourcelines(fn)[1]
    except (TypeError, OSError):
        return -1


def merge_doc(out, doc):
    doc, params, ret = parse_rst(doc)
    out["doc"] = doc

    for param in params:
        for out_param in out["params"]:
            if out_param["name"] != param["name"]:
                continue
            out_param["doc"] = param["doc"] or out_param["doc"]
            out_param["type"] = param["type"] or out_param["type"]
            break

    out["return"]["doc"] = ret["doc"] or out["return"]["doc"]
    out["return"]["type"] = ret["type"] or out["return"]["type"]
    return out


def rst_read_doc(lines):
    doc = []
    for i, line in enumerate(lines):
        if line[:1] == ":":
            return i, "\n".join(doc).strip()
        doc.append(line)
    return -1, "\n".join(doc).strip()


def rst_read_section(lines, i):
    # Skip empty lines/other lines
    for i, line in enumerate(lines[i:], i):
        if not line.strip():
            continue
        # :param path: The path of the file to wrap
        match = re.match(r":\s*(\w+)(\s+\w+)?\s*:", lines[i])
        if match:
            break
    else:
        return None

    tag = match.group(1)
    value = match.group(2).strip() if match.group(2) else ""
    text = lines[i][match.end() :].lstrip()
    for i in range(i + 1, len(lines)):
        if re.match(r"\t+| {3,}", lines[i]):
            text += " " + lines[i].lstrip()
        else:
            return tag, value, text, i
    return tag, value, text.strip(), -1


def parse_rst(docstring: str):
    lines = docstring.splitlines()
    i, doc = rst_read_doc(lines)
    params, names = {}, []
    ret = param_dict()

    while i != -1:
        out = rst_read_section(lines, i)
        if not out:
            break
        tag, value, text, i = out
        if tag == "param":
            params[value] = param_dict(name=value, doc=text)
            names.append(value)
        elif tag == "type":
            # TODO: Check param
            params[value]["type"] = text
        elif tag == "returns":
            ret["doc"] = text
        elif tag == "rtype":
            ret["type"] = text
        else:
            raise ValueError(f"{i+1}: unknown tag - {lines[i]!r}")

    params = [params[name] for name in names]
    return doc, params, ret


def ast_func_info(func: ast.FunctionDef):
    doc = ast.get_docstring(func) or ""
    rtype = None
    if func.returns:
        rtype = ast.unparse(func.returns)
    params = [ast_param_dict(p) for p in func.args.args]
    # adds info about *args and **kwargs to the function doc
    has_varargs = func.args.vararg is not None
    has_kwargs = func.args.kwarg is not None
    defaults = func.args.defaults
    if defaults:
        for param, default in zip(params[-len(defaults) :], defaults):
            param["default"] = ast_code(default)

    out = func_dict(
        name=func.name,
        doc=doc,
        params=params,
        returns=param_dict(type=rtype, default=None),
        lineno=func.lineno,
        has_varargs=has_varargs,
        has_kwargs=has_kwargs,
    )

    if not doc.strip():
        return out

    return merge_doc(out, doc)


def ast_param_dict(param: ast.arg) -> dict:
    return {
        "name": param.arg,
        "type": ann_type(param.annotation) if param.annotation else "",
        "doc": "",
    }


def ann_type(ann):
    if hasattr(ann, "slice"):
        if isinstance(ann.value, ast.Attribute):
            # value is an attribute, e.g. b of a.b - get the full path
            name = get_attr_path(ann.value)
        else:
            name = ann.value.id
        inner = ", ".join(ann_type(e) for e in iter_elems(ann.slice))
        return f"{name}[{inner}]"

    if isinstance(ann, ast.Attribute):
        if isinstance(ann.value, ast.Attribute):
            # value is an attribute, e.g. b of a.b - get the full path
            return get_attr_path(ann)

        return ann.attr

    return getattr(ann, "id", "")


def get_attr_path(ann: ast.Attribute):
    if isinstance(ann.value, ast.Attribute):
        # value is an attribute, e.g. b of a.b - get the full path
        return f"{get_attr_path(ann.value)}.{ann.attr}"

    # value can be a subscript or name - get its annotation type and append the attribute
    return f"{ann_type(ann.value)}.{ann.attr}"


def iter_elems(ann):
    """
    Gets the elements of an ast.Subscript.slice, e.g. Union[int, str] -> [int, str]
    """
    if hasattr(ann, "elts"):
        return ann.elts
    if hasattr(ann, "id"):
        return [ann]
    if hasattr(ann.value, "elts"):
        return ann.value.elts
    if not hasattr(ann, "slice"):
        return [ann.value]

    return [ann]


class ASTVisitor(ast.NodeVisitor):
    def __init__(self):
        self.funcs = []
        self.exprs = []

    def generic_visit(self, node):
        self.exprs.append(node)
        super().generic_visit(node)

    def visit_FunctionDef(self, node):  # noqa: N802
        self.funcs.append(node)
        self.generic_visit(node)


def find_handlers(code: str, handlers=None):
    handlers = set() if handlers is None else set(handlers)
    mod = ast.parse(code)
    visitor = ASTVisitor()
    visitor.visit(mod)
    funcs = [ast_func_info(fn) for fn in visitor.funcs]
    if handlers:
        return [f for f in funcs if f["name"] in handlers]
    else:
        markers = find_handler_markers(code)
        return filter_funcs(funcs, markers)


def as_func(handler):
    ret = clean(handler["return"])
    return FunctionEntrypoint(
        name=handler["name"],
        doc=handler["doc"],
        parameters=[clean(p) for p in handler["params"]],
        outputs=[ret] if ret else None,
        lineno=handler["lineno"],
        has_varargs=handler["has_varargs"],
        has_kwargs=handler["has_kwargs"],
    ).to_dict()


def update_function_entry_points(function, code):
    handlers = find_handlers(code)
    function.spec.entry_points = {
        handler["name"]: as_func(handler) for handler in handlers
    }


def clean(struct: dict):
    if not struct:
        return None
    if "default" in struct:
        struct["default"] = py_eval(struct["default"])
    return {k: v for k, v in struct.items() if v or k == "default"}


def py_eval(data):
    try:
        value = ast.literal_eval(data)
        return value
    except (SyntaxError, ValueError):
        return data


def filter_funcs(funcs, markers):
    markers = list(markers)
    if not markers:
        return [func for func in funcs if func["name"][0] != "_"]

    return [func for func in funcs if is_marked(func, markers)]


def is_marked(func, markers):
    for marker in markers:
        if func["lineno"] - marker == 1:
            return True
    return False


def find_handler_markers(code: str):
    for lnum, line in enumerate(code.splitlines(), 1):
        # '# mlrun:handler'
        if re.match(r"#\s*mlrun:handler", line):
            yield lnum


def ast_code(expr):
    # Sadly, not such built in
    children = None
    if isinstance(expr, ast.Dict):
        children = zip(expr.keys, expr.values)
        children = [f"{ast_code(k)}: {ast_code(v)}" for k, v in children]
        inner = ", ".join(children)
        return f"{{{inner}}}"
    elif isinstance(expr, ast.Set):
        start, end, children = "{", "}", expr.elts
        if not children:
            return "set()"
    elif isinstance(expr, ast.Tuple):
        start, end, children = "(", ")", expr.elts
    elif isinstance(expr, ast.List):
        start, end, children = "[", "]", expr.elts
    elif isinstance(expr, ast.Call):
        children = [ast_code(e) for e in expr.args]
        children += [f"{k.arg}={ast_code(k.value)}" for k in expr.keywords]
        inner = ", ".join(children)
        return f"{expr.func.id}({inner})"
    else:  # Leaf (number, str ...)
        return repr(getattr(expr, expr._fields[0]))

    inner = ", ".join(ast_code(e) for e in children)
    return f"{start}{inner}{end}"



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextvars
import datetime
import logging
import os
import string
import sys
import typing
from enum import Enum
from functools import cached_property
from sys import stdout
from traceback import format_exception
from typing import IO, Optional, Union

import orjson
import pydantic.v1

from mlrun import errors
from mlrun.config import config

context_id_var = contextvars.ContextVar("context_id", default=None)


class _BaseFormatter(logging.Formatter):
    def _json_dump(self, json_object):
        def default(obj):
            if isinstance(obj, pydantic.v1.BaseModel):
                return obj.dict()

            # EAFP all the way.
            # Leave the unused "exc" in for debugging ease
            try:
                return obj.__log__()
            except Exception as exc:  # noqa
                try:
                    return obj.__repr__()
                except Exception as exc:  # noqa
                    try:
                        return str(obj)
                    except Exception as exc:
                        raise TypeError from exc

        return orjson.dumps(
            json_object,
            option=orjson.OPT_NAIVE_UTC
            | orjson.OPT_SERIALIZE_NUMPY
            | orjson.OPT_NON_STR_KEYS
            | orjson.OPT_SORT_KEYS,
            default=default,
        ).decode()

    def _record_with(self, record):
        record_with = getattr(record, "with", {})
        if record.exc_info:
            record_with.update(exc_info=format_exception(*record.exc_info))
        if "ctx" not in record_with:
            if (ctx_id := context_id_var.get()) is not None:
                record_with["ctx"] = ctx_id
        return record_with


class JSONFormatter(_BaseFormatter):
    def format(self, record) -> str:
        record_with = self._record_with(record)
        record_fields = {
            "datetime": self.formatTime(record, self.datefmt),
            "level": record.levelname.lower(),
            "message": record.getMessage(),
            "with": record_with,
        }

        return self._json_dump(record_fields)


class HumanReadableFormatter(_BaseFormatter):
    def format(self, record) -> str:
        more = self._resolve_more(record)
        return (
            f"> {self.formatTime(record, self.datefmt)} "
            f"[{record.levelname.lower()}] "
            f"{record.getMessage().rstrip()}"
            f"{more}"
        )

    def _resolve_more(self, record):
        record_with = self._record_with(record)
        record_with_encoded = self._json_dump(record_with) if record_with else ""
        more = f": {record_with_encoded}" if record_with_encoded else ""
        return more


class CustomFormatter(HumanReadableFormatter):
    """
    To enable custom logger formatter, configure MLRun with the following env variables:
    1. "MLRUN_LOG_FORMATTER" = "custom" - change the default log formatter.
    2. "MLRUN_LOG_FORMAT_OVERRIDE" = "> {timestamp} [{level}] Running module: {module} {message} {more}" - logger format
        * Note that your custom format must include those 4 fields - timestamp, level, message and more
    If the custom format is not configured properly , MLRun will use the default logger (human format).
    """

    # This attribute is used to solve an issue
    # that causes the warning to be written numerous times(for any log generation).
    # We want to print the errors just once, not for each logger generation.
    fail_on_format_configuration = False  # for issues that relates to unrecognized keys
    fail_on_missing_default_keys_key = (
        False  # for issues that relates to missing default keys
    )

    def format(self, record) -> str:
        more = self._resolve_more(record)
        custom_format = config.log_format_override
        _custom_format = None
        current_time = datetime.datetime.now()
        formatted_time = current_time.strftime("%Y-%m-%d %H:%M:%S,%f")[:-3]
        try:
            if custom_format:
                default_keys = ["timestamp", "level", "message", "more"]
                formatter = string.Formatter()
                custom_format_keys = [
                    key
                    for _, key, _, _ in formatter.parse(custom_format)
                    if key is not None
                ]
                missing_default_flags = list(
                    set(default_keys) - set(custom_format_keys)
                )

                if (
                    missing_default_flags
                    and not CustomFormatter.fail_on_missing_default_keys_key
                ):
                    print(
                        f'> {formatted_time} [warning] Custom loggers must '
                        f'include those keys within the logger format, {", ".join(default_keys)} '
                        f'your format is missing: {", ".join(missing_default_flags)}',
                        file=sys.stderr,
                    )
                    CustomFormatter.fail_on_missing_default_keys_key = True
                record_dict = record.__dict__
                missing_format_configuraiton_keys = list(
                    set(custom_format_keys)
                    - set(default_keys)
                    - set(record_dict.keys())
                )
                if missing_format_configuraiton_keys:
                    if not CustomFormatter.fail_on_format_configuration:
                        print(
                            f"> {formatted_time} [warning] Failed to create custom logger due "
                            f'to missing format key in the log record: {", ".join(missing_format_configuraiton_keys)}',
                            file=sys.stderr,
                        )
                        CustomFormatter.fail_on_format_configuration = True
                    _format = (
                        f"> {self.formatTime(record, self.datefmt)} "
                        f"[{record.levelname.lower()}] "
                        f"{record.getMessage().rstrip()}"
                        f"{more}"
                    )
                _custom_format = custom_format.format(
                    timestamp=self.formatTime(record, self.datefmt),
                    level=record.levelname.lower(),
                    message=record.getMessage().rstrip(),
                    more=more or "",
                    **record_dict,
                )
                CustomFormatter.fail_on_format_configuration = True
        except Exception as e:
            if not CustomFormatter.fail_on_format_configuration:
                print(
                    f"> {formatted_time} [warning] Failed to create custom logger, "
                    f"see Exception: {errors.err_to_str(e)}",
                    file=sys.stderr,
                )
                CustomFormatter.fail_on_format_configuration = True
        _format = _custom_format or (
            f"> {self.formatTime(record, self.datefmt)} "
            f"[{record.levelname.lower()}] "
            f"{record.getMessage().rstrip()}"
            f"{more}"
        )
        return _format


class HumanReadableExtendedFormatter(HumanReadableFormatter):
    _colors = {
        logging.NOTSET: "",
        logging.DEBUG: "\x1b[34m",
        logging.INFO: "\x1b[36m",
        logging.WARNING: "\x1b[33m",
        logging.ERROR: "\x1b[0;31m",
        logging.CRITICAL: "\x1b[1;31m",
    }
    _color_reset = "\x1b[0m"

    def format(self, record) -> str:
        more = ""
        record_with = self._record_with(record)
        if record_with:

            def _format_value(val):
                formatted_val = (
                    val
                    if isinstance(val, str)
                    else str(orjson.loads(self._json_dump(val)))
                )
                return (
                    formatted_val.replace("\n", "\n\t\t")
                    if len(formatted_val) < 4096
                    else repr(formatted_val)
                )

            more = "\n\t" + "\n\t".join(
                [f"{key}: {_format_value(val)}" for key, val in record_with.items()]
            )
        return (
            f"{self._get_message_color(record.levelno)}> "
            f"{self.formatTime(record, self.datefmt)} "
            f"[{record.name}:{record.levelname.lower()}] "
            f"{record.getMessage()}{more}{self._get_color_reset()}"
        )

    def _get_color_reset(self):
        if not self._have_color_support:
            return ""

        return self._color_reset

    def _get_message_color(self, levelno):
        if not self._have_color_support:
            return ""

        return self._colors[levelno]

    @cached_property
    def _have_color_support(self):
        if os.environ.get("PYCHARM_HOSTED"):
            return True
        if os.environ.get("NO_COLOR"):
            return False
        if os.environ.get("CLICOLOR_FORCE"):
            return True
        return stdout.isatty()


class Logger:
    def __init__(
        self,
        level,
        name="mlrun",
        propagate=True,
        logger: Optional[logging.Logger] = None,
    ):
        self._logger = logger or logging.getLogger(name)
        self._logger.propagate = propagate
        self._logger.setLevel(level)
        self._bound_variables = {}

        for log_level_func in [
            self.exception,
            self.error,
            self.warn,
            self.warning,
            self.info,
            self.debug,
        ]:
            setattr(self, f"{log_level_func.__name__}_with", log_level_func)

    def set_handler(
        self, handler_name: str, file: IO[str], formatter: logging.Formatter
    ):
        # check if there's a handler by this name
        for handler in self._logger.handlers:
            if handler.name == handler_name:
                self._logger.removeHandler(handler)
                break

        # create a stream handler from the file
        stream_handler = logging.StreamHandler(file)
        stream_handler.name = handler_name

        # set the formatter
        stream_handler.setFormatter(formatter)

        # add the handler to the logger
        self._logger.addHandler(stream_handler)

    def get_child(self, suffix):
        """
        Get a child logger with the given suffix.
        This is useful for when you want to have a logger for a specific component.
        Once the formatter will support logger name, it will be easier to understand
        which component logged the message.

        :param suffix: The suffix to add to the logger name.
        """
        return Logger(
            self.level,
            # name is not set as it is provided by the "getChild"
            name="",
            # allowing child to delegate events logged to ancestor logger
            # not doing so, will leave log lines not being handled
            propagate=True,
            logger=self._logger.getChild(suffix),
        )

    @property
    def level(self):
        return self._logger.level

    def set_logger_level(self, level: Union[str, int]) -> None:
        self._logger.setLevel(level)

    def replace_handler_stream(self, handler_name: str, file: IO[str]) -> None:
        self.get_handler(handler_name).stream = file

    def get_handler(self, name: str) -> logging.Handler:
        for handler in self._logger.handlers:
            if handler.name == name:
                return handler
        raise ValueError(f"Logger does not have a handler named '{name}'")

    def debug(self, message, *args, **kw_args):
        self._update_bound_vars_and_log(logging.DEBUG, message, *args, **kw_args)

    def info(self, message, *args, **kw_args):
        self._update_bound_vars_and_log(logging.INFO, message, *args, **kw_args)

    def warn(self, message, *args, **kw_args):
        self._update_bound_vars_and_log(logging.WARNING, message, *args, **kw_args)

    def warning(self, message, *args, **kw_args):
        self.warn(message, *args, **kw_args)

    def error(self, message, *args, **kw_args):
        self._update_bound_vars_and_log(logging.ERROR, message, *args, **kw_args)

    def exception(self, message, *args, exc_info=True, **kw_args):
        self._update_bound_vars_and_log(
            logging.ERROR, message, *args, exc_info=exc_info, **kw_args
        )

    def bind(self, **kw_args):
        self._bound_variables.update(kw_args)

    def _update_bound_vars_and_log(
        self, level, message, *args, exc_info=None, **kw_args
    ):
        kw_args.update(self._bound_variables)
        if kw_args:
            self._logger.log(
                level, message, *args, exc_info=exc_info, extra={"with": kw_args}
            )
            return

        self._logger.log(level, message, *args, exc_info=exc_info)


class FormatterKinds(Enum):
    HUMAN = "human"
    HUMAN_EXTENDED = "human_extended"
    JSON = "json"
    CUSTOM = "custom"


def resolve_formatter_by_kind(
    formatter_kind: FormatterKinds,
) -> type[
    typing.Union[
        HumanReadableFormatter,
        HumanReadableExtendedFormatter,
        JSONFormatter,
        CustomFormatter,
    ]
]:
    return {
        FormatterKinds.HUMAN: HumanReadableFormatter,
        FormatterKinds.HUMAN_EXTENDED: HumanReadableExtendedFormatter,
        FormatterKinds.JSON: JSONFormatter,
        FormatterKinds.CUSTOM: CustomFormatter,
    }[formatter_kind]


def create_test_logger(name: str = "mlrun", stream: IO[str] = stdout) -> Logger:
    logger = create_logger(
        level="debug",
        formatter_kind=FormatterKinds.HUMAN_EXTENDED.name,
        name=name,
        stream=stream,
    )
    logger._logger.propagate = True  # pass records up to pytest’s handler
    return logger


def create_logger(
    level: Optional[str] = None,
    formatter_kind: str = FormatterKinds.HUMAN.name,
    name: str = "mlrun",
    stream: IO[str] = stdout,
) -> Logger:
    level = level or config.log_level or "info"

    level = logging.getLevelName(level.upper())

    # create logger instance
    logger_instance = Logger(level, name=name, propagate=False)

    # resolve formatter
    formatter_instance = resolve_formatter_by_kind(
        FormatterKinds(formatter_kind.lower())
    )

    # set handler
    logger_instance.set_handler("default", stream or stdout, formatter_instance())

    return logger_instance



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mlrun.errors import MLRunBaseError


class MLRunPackageError(MLRunBaseError):
    """
    General error from `mlrun.package`.
    """

    pass


class MLRunPackageCollectionError(MLRunPackageError):
    """
    An error that may be raised during the collection of packagers the manager is assigned to do.
    """

    pass


class MLRunPackagePackingError(MLRunPackageError):
    """
    An error that may be raised during a `mlrun.Packager.pack` method.
    """

    pass


class MLRunPackageUnpackingError(MLRunPackageError):
    """
    An error that may be raised during a `mlrun.Packager.unpack` method.
    """

    pass



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
from typing import Any

import mlrun.feature_store.steps
from mlrun.common.schemas.model_monitoring import (
    EventFieldType,
    EventKeyMetrics,
    EventLiveStats,
)
from mlrun.utils import logger


def _normalize_dict_for_v3io_frames(event: dict[str, Any]) -> dict[str, Any]:
    """
    Normalize user-defined keys (e.g., model input data and predictions) to a format V3IO Frames tolerates.

    - Keys must match regex: '^[a-zA-Z_:]([a-zA-Z0-9_:])*$'
    - Replace invalid characters (e.g., '-') with '_'.
    - Prefix keys starting with digits with '_'.
    - Flatten nested dictionaries using dot notation, while normalizing keys recursively.
    """
    prefix = "_"

    def norm_key(key: str) -> str:
        key = key.replace("-", "_")  # hyphens `-` are not allowed
        if key and key[0].isdigit():  # starting with a digit is not allowed
            return prefix + key
        return key

    def flatten_dict(d: dict[str, Any], parent_key: str = "") -> dict[str, Any]:
        items = {}
        for k, v in d.items():
            new_key = norm_key(k)
            full_key = f"{parent_key}.{new_key}" if parent_key else new_key
            if isinstance(v, dict):
                items.update(flatten_dict(v, full_key))
            else:
                items[full_key] = v
        return items

    return flatten_dict(event)


class ProcessBeforeTSDB(mlrun.feature_store.steps.MapClass):
    def __init__(self, **kwargs):
        """
        Process the data before writing to TSDB. This step creates a dictionary that includes 3 different dictionaries
        that each one of them contains important details and stats about the events:
        1. base_metrics: stats about the average latency and the amount of predictions over time. It is based on
           storey.AggregateByKey which was executed in step 5.
        2. endpoint_features: feature names and values along with the prediction names and value.
        3. custom_metric (opt): optional metrics provided by the user.
        :returns: Dictionary of 2-3 dictionaries that contains stats and details about the events.
        """
        super().__init__(**kwargs)

    def do(self, event):
        # Compute prediction per second
        event[EventLiveStats.PREDICTIONS_PER_SECOND] = (
            float(event[EventLiveStats.PREDICTIONS_COUNT_5M]) / 300
        )
        base_fields = [
            EventFieldType.TIMESTAMP,
            EventFieldType.ENDPOINT_ID,
            EventFieldType.ENDPOINT_TYPE,
        ]

        # Getting event timestamp and endpoint_id
        base_event = {k: event[k] for k in base_fields}

        # base_metrics includes the stats about the average latency and the amount of predictions over time
        base_metrics = {
            EventFieldType.RECORD_TYPE: EventKeyMetrics.BASE_METRICS,
            EventLiveStats.PREDICTIONS_PER_SECOND: event[
                EventLiveStats.PREDICTIONS_PER_SECOND
            ],
            EventLiveStats.PREDICTIONS_COUNT_5M: event[
                EventLiveStats.PREDICTIONS_COUNT_5M
            ],
            EventLiveStats.PREDICTIONS_COUNT_1H: event[
                EventLiveStats.PREDICTIONS_COUNT_1H
            ],
            EventLiveStats.LATENCY_AVG_5M: event[EventLiveStats.LATENCY_AVG_5M],
            EventLiveStats.LATENCY_AVG_1H: event[EventLiveStats.LATENCY_AVG_1H],
            **base_event,
        }

        # endpoint_features includes the event values of each feature and prediction
        endpoint_features = {
            EventFieldType.RECORD_TYPE: EventKeyMetrics.ENDPOINT_FEATURES,
            **_normalize_dict_for_v3io_frames(event[EventFieldType.NAMED_PREDICTIONS]),
            **_normalize_dict_for_v3io_frames(event[EventFieldType.NAMED_FEATURES]),
            **base_event,
        }
        # Create a dictionary that includes both base_metrics and endpoint_features
        processed = {
            EventKeyMetrics.BASE_METRICS: base_metrics,
            EventKeyMetrics.ENDPOINT_FEATURES: endpoint_features,
        }

        # If metrics provided, add another dictionary if custom_metrics values
        if event[EventFieldType.METRICS]:
            processed[EventKeyMetrics.CUSTOM_METRICS] = {
                EventFieldType.RECORD_TYPE: EventKeyMetrics.CUSTOM_METRICS,
                **event[EventFieldType.METRICS],
                **base_event,
            }

        return processed


class FilterAndUnpackKeys(mlrun.feature_store.steps.MapClass):
    def __init__(self, keys, **kwargs):
        """
        Create unpacked event dictionary based on provided key metrics (base_metrics, endpoint_features,
        or custom_metric). Please note that the next step of the TSDB target requires an unpacked dictionary.
        :param keys: list of key metrics.
        :returns: An unpacked dictionary of event filtered by the provided key metrics.
        """
        super().__init__(**kwargs)
        self.keys = keys

    def do(self, event):
        # Keep only the relevant dictionary based on the provided keys
        new_event = {}
        for key in self.keys:
            if key in event:
                new_event[key] = event[key]

        # Create unpacked dictionary
        unpacked = {}
        for key in new_event.keys():
            if key in self.keys:
                unpacked = {**unpacked, **new_event[key]}
            else:
                unpacked[key] = new_event[key]
        return unpacked if unpacked else None


class ErrorExtractor(mlrun.feature_store.steps.MapClass):
    def __init__(self, **kwargs):
        """
        Prepare the event for insertion into the errors TSDB table.
        """
        super().__init__(**kwargs)

    def do(self, event):
        error = event.get("error")
        timestamp = datetime.fromisoformat(event.get("when"))
        endpoint_id = event[EventFieldType.ENDPOINT_ID]
        event = {
            EventFieldType.MODEL_ERROR: str(error),
            EventFieldType.ERROR_TYPE: EventFieldType.INFER_ERROR,
            EventFieldType.ENDPOINT_ID: endpoint_id,
            EventFieldType.TIMESTAMP: timestamp,
            EventFieldType.ERROR_COUNT: 1.0,
        }
        logger.info("Write error to errors TSDB table", event=event)
        return event



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import re
import uuid
from collections import OrderedDict
from typing import Any, Optional, Union

import numpy as np
import pandas as pd
from storey import MapClass

import mlrun.errors
from mlrun.serving.utils import StepToDict
from mlrun.utils import get_in


def get_engine(first_event):
    if hasattr(first_event, "body"):
        first_event = first_event.body
    if isinstance(first_event, pd.DataFrame):
        return "pandas"
    if hasattr(first_event, "rdd"):
        return "spark"
    return "storey"


class MLRunStep(MapClass):
    def __init__(self, **kwargs):
        """Abstract class for mlrun step.
        Can be used in pandas/storey/spark feature set ingestion. Extend this class and implement the relevant
        `_do_XXX` methods to support the required execution engines.
        """
        super().__init__(**kwargs)
        self._engine_to_do_method = {
            "pandas": self._do_pandas,
            "spark": self._do_spark,
            "storey": self._do_storey,
        }

    def do(self, event):
        """
        This method defines the do method of this class according to the first event type.

        .. warning::
            When extending this class, do not override this method; only override the `_do_XXX` methods.
        """
        engine = get_engine(event)
        self.do = self._engine_to_do_method.get(engine, None)
        if self.do is None:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Unrecognized engine: {engine}. Available engines are: pandas, spark and storey"
            )

        return self.do(event)

    def _do_pandas(self, event):
        """
        The execution method for pandas engine.

        :param event: Incoming event, a `pandas.DataFrame` object.
        """
        raise NotImplementedError

    def _do_storey(self, event):
        """
        The execution method for storey engine.

        :param event: Incoming event, a dictionary or `storey.Event` object, depending on the `full_event` value.
        """
        raise NotImplementedError

    def _do_spark(self, event):
        """
        The execution method for spark engine.

        :param event: Incoming event, a `pyspark.sql.DataFrame` object.
        """
        raise NotImplementedError


class FeaturesetValidator(StepToDict, MLRunStep):
    def __init__(self, featureset=None, columns=None, name=None, **kwargs):
        """Validate feature values according to the feature set validation policy

        :param featureset: feature set uri (or "." for current feature set pipeline)
        :param columns:    names of the columns/fields to validate
        :param name:       step name
        :param kwargs:     optional kwargs (for storey)
        """
        kwargs["full_event"] = True
        super().__init__(**kwargs)
        self._validators = {}
        self.featureset = featureset or "."
        self.columns = columns
        self.name = name
        if not self.context:
            return
        self._featureset = self.context.get_store_resource(featureset)
        for key, feature in self._featureset.spec.features.items():
            if feature.validator and (not columns or key in columns):
                feature.validator.set_feature(feature)
                self._validators[key] = feature.validator

    def _do_storey(self, event):
        body = event.body
        for name, validator in self._validators.items():
            if name in body:
                ok, args = validator.check(body[name])
                if not ok:
                    message = args.pop("message")
                    key_text = f" key={event.key}" if event.key else ""
                    print(
                        f"{validator.severity}! {name} {message},{key_text} args={args}"
                    )
        return event

    def _do_pandas(self, event):
        body = event.body
        for column in body:
            validator = self._validators.get(column, None)
            if validator:
                violations = 0
                all_args = []
                for i in body.index:
                    # check each body entry if there is validator for it
                    ok, args = validator.check(body.at[i, column])
                    if not ok:
                        violations += 1
                        all_args.append(args)
                        message = args.pop("message")
                if violations != 0:
                    text = f" column={column}, has {violations} violations"
                    print(
                        f"{validator.severity}! {column} {message},{text} args={all_args}"
                    )
        return event


class MapValues(StepToDict, MLRunStep):
    def __init__(
        self,
        mapping: dict[str, dict[Union[str, int, bool], Any]],
        with_original_features: bool = False,
        suffix: str = "mapped",
        **kwargs,
    ):
        """Map column values to new values

        example::

            # replace the value "U" with '0' in the age column
            graph.to(MapValues(mapping={"age": {"U": "0"}}, with_original_features=True))

            # replace integers, example
            graph.to(MapValues(mapping={"not": {0: 1, 1: 0}}))

            # replace by range, use -inf and inf for extended range
            graph.to(
                MapValues(
                    mapping={
                        "numbers": {"ranges": {"negative": [-inf, 0], "positive": [0, inf]}}
                    }
                )
            )

        :param mapping: a dict with entry per column and the associated old/new values map
        :param with_original_features: set to True to keep the original features
        :param suffix: the suffix added to the column name <column>_<suffix> (default is "mapped")
        :param kwargs: optional kwargs (for storey)
        """
        super().__init__(**kwargs)
        self.mapping = mapping
        self.with_original_features = with_original_features
        self.suffix = suffix

    def _map_value(self, feature: str, value):
        feature_map = self.mapping.get(feature, {})

        # Is this a range replacement?
        if "ranges" in feature_map:
            for val, val_range in feature_map.get("ranges", {}).items():
                min_val = val_range[0] if val_range[0] != "-inf" else -np.inf
                max_val = val_range[1] if val_range[1] != "inf" else np.inf
                if value >= min_val and value < max_val:
                    return val

        # Is it a regular replacement
        return feature_map.get(value, value)

    def _get_feature_name(self, feature) -> str:
        return f"{feature}_{self.suffix}" if self.with_original_features else feature

    def _do_storey(self, event):
        mapped_values = {
            self._get_feature_name(feature): self._map_value(feature, val)
            for feature, val in event.items()
            if feature in self.mapping
        }

        if self.with_original_features:
            mapped_values.update(event)

        return mapped_values

    def _do_pandas(self, event):
        df = pd.DataFrame(index=event.index)
        for feature in event.columns:
            feature_map = self.mapping.get(feature, {})
            if "ranges" in feature_map:
                # create and apply range map
                for val, val_range in feature_map.get("ranges", {}).items():
                    min_val = val_range[0] if val_range[0] != "-inf" else -np.inf
                    max_val = val_range[1] if val_range[1] != "inf" else np.inf
                    feature_map["ranges"][val] = [min_val, max_val]

                matchdf = pd.DataFrame.from_dict(
                    feature_map["ranges"], "index"
                ).reset_index()
                matchdf.index = pd.IntervalIndex.from_arrays(
                    left=matchdf[0], right=matchdf[1], closed="both"
                )
                df[self._get_feature_name(feature)] = matchdf.loc[event[feature]][
                    "index"
                ].values
            elif feature_map:
                # create and apply simple map
                df[self._get_feature_name(feature)] = event[feature].map(
                    lambda x: feature_map.get(x, None)
                )

        if self.with_original_features:
            df = pd.concat([event, df], axis=1)
        return df

    def _do_spark(self, event):
        from itertools import chain

        from pyspark.sql.functions import col, create_map, isnan, isnull, lit, when
        from pyspark.sql.types import DecimalType, DoubleType, FloatType
        from pyspark.sql.utils import AnalysisException

        df = event
        source_column_names = df.columns
        for column, column_map in self.mapping.items():
            new_column_name = self._get_feature_name(column)
            if self.get_ranges_key() not in column_map:
                if column not in source_column_names:
                    continue
                mapping_expr = create_map([lit(x) for x in chain(*column_map.items())])
                try:
                    df = df.withColumn(
                        new_column_name,
                        when(
                            col(column).isin(list(column_map.keys())),
                            mapping_expr.getItem(col(column)),
                        ).otherwise(col(column)),
                    )
                #  if failed to use otherwise it is probably because the new column has different type
                #  then the original column.
                #  we will try to replace the values without using 'otherwise'.
                except AnalysisException:
                    df = df.withColumn(
                        new_column_name, mapping_expr.getItem(col(column))
                    )
                    col_type = df.schema[column].dataType
                    new_col_type = df.schema[new_column_name].dataType
                    #  in order to avoid exception at isna on non-decimal/float columns -
                    #  we need to check their types before filtering.
                    if isinstance(col_type, (FloatType, DoubleType, DecimalType)):
                        column_filter = (~isnull(col(column))) & (~isnan(col(column)))
                    else:
                        column_filter = ~isnull(col(column))
                    if isinstance(new_col_type, (FloatType, DoubleType, DecimalType)):
                        new_column_filter = isnull(col(new_column_name)) | isnan(
                            col(new_column_name)
                        )
                    else:
                        #  we need to check that every value replaced if we changed column type - except None or NaN.
                        new_column_filter = isnull(col(new_column_name))
                    mapping_to_null = [
                        k
                        for k, v in column_map.items()
                        if v is None
                        or (
                            isinstance(v, (float, np.float64, np.float32, np.float16))
                            and math.isnan(v)
                        )
                    ]
                    turned_to_none_values = df.filter(
                        column_filter & new_column_filter
                    ).filter(~col(column).isin(mapping_to_null))

                    if len(turned_to_none_values.head(1)) > 0:
                        raise mlrun.errors.MLRunInvalidArgumentError(
                            f"MapValues - mapping that changes column type must change all values accordingly,"
                            f" which is not the case for column '{column}'"
                        )
            else:
                for val, val_range in column_map["ranges"].items():
                    min_val = val_range[0] if val_range[0] != "-inf" else -np.inf
                    max_val = val_range[1] if val_range[1] != "inf" else np.inf
                    otherwise = ""
                    if new_column_name in df.columns:
                        otherwise = df[new_column_name]
                    df = df.withColumn(
                        new_column_name,
                        when(
                            (df[column] < max_val) & (df[column] >= min_val),
                            lit(val),
                        ).otherwise(otherwise),
                    )

        if not self.with_original_features:
            df = df.select(*self.mapping.keys())

        return df

    @classmethod
    def validate_args(cls, feature_set, **kwargs):
        mapping = kwargs.get("mapping", [])
        for column, column_map in mapping.items():
            if cls.get_ranges_key() not in column_map:
                types = set(
                    type(val)
                    for val in column_map.values()
                    if type(val) is not None
                    and not (
                        isinstance(val, (float, np.float64, np.float32, np.float16))
                        and math.isnan(val)
                    )
                )
            else:
                if len(column_map) > 1:
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        f"MapValues - mapping values of the same column can not combine ranges and "
                        f"single replacement, which is the case for column '{column}'"
                    )
                ranges_dict = column_map[cls.get_ranges_key()]
                types = set()
                for ranges_mapping_values in ranges_dict.values():
                    range_types = set(
                        type(val)
                        for val in ranges_mapping_values
                        if type(val) is not None
                        and val != "-inf"
                        and val != "inf"
                        and not (
                            isinstance(val, (float, np.float64, np.float32, np.float16))
                            and math.isnan(val)
                        )
                    )
                    types = types.union(range_types)
            if len(types) > 1:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"MapValues - mapping values of the same column must be in the"
                    f" same type, which was not the case for Column '{column}'"
                )

    @staticmethod
    def get_ranges_key():
        return "ranges"


class Imputer(StepToDict, MLRunStep):
    def __init__(
        self,
        method: str = "avg",
        default_value=None,
        mapping: Optional[dict[str, Any]] = None,
        **kwargs,
    ):
        """Replace None values with default values

        :param method:        for future use
        :param default_value: default value if not specified per column
        :param mapping:       a dict of per column default value
        :param kwargs:        optional kwargs (for storey)
        """
        super().__init__(**kwargs)
        self.mapping = mapping or {}
        self.method = method
        self.default_value = default_value

    def _impute(self, feature: str, value: Any):
        if pd.isna(value):
            return self.mapping.get(feature, self.default_value)
        return value

    def _do_storey(self, event):
        imputed_values = {
            feature: self._impute(feature, val) for feature, val in event.items()
        }
        return imputed_values

    def _do_pandas(self, event):
        for feature in event.columns:
            val = self.mapping.get(feature, self.default_value)
            if val is not None:
                event[feature].fillna(val, inplace=True)
        return event

    def _do_spark(self, event):
        for feature in event.columns:
            val = self.mapping.get(feature, self.default_value)
            if val is not None:
                event = event.na.fill(val, feature)
                # for future use - for now sparks=storey=pandas
                # from pyspark.ml.feature import Imputer
                # imputer = Imputer(inputCols=[feature], outputCols=[feature]).setStrategy(val)
                # event = imputer.fit(event).transform(event)
        return event


class OneHotEncoder(StepToDict, MLRunStep):
    def __init__(self, mapping: dict[str, list[Union[int, str]]], **kwargs):
        """Create new binary fields, one per category (one hot encoded)

        example::

            mapping = {
                "category": ["food", "health", "transportation"],
                "gender": ["male", "female"],
            }
            graph.to(OneHotEncoder(mapping=one_hot_encoder_mapping))

        :param mapping: a dict of per column categories (to map to binary fields)
        :param kwargs:  optional kwargs (for storey)
        """
        super().__init__(**kwargs)
        self.mapping = mapping
        for key, values in mapping.items():
            for val in values:
                if not (isinstance(val, str) or isinstance(val, (int, np.integer))):
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        "For OneHotEncoder you must provide int or string mapping list"
                    )
            # Use OrderedDict to dedup without losing the original order
            mapping[key] = list(OrderedDict.fromkeys(values).keys())

    def _encode(self, feature: str, value):
        encoding = self.mapping.get(feature, [])

        if encoding:
            one_hot_encoding = {
                f"{feature}_{OneHotEncoder._sanitized_category(category)}": 0
                for category in encoding
            }
            if value in encoding:
                one_hot_encoding[
                    f"{feature}_{OneHotEncoder._sanitized_category(value)}"
                ] = 1
            elif self.logger:
                self.logger.warn(
                    f"OneHotEncoder does not have an encoding for value '{value}' of feature '{feature}'"
                )
            return one_hot_encoding

        return {feature: value}

    def _do_storey(self, event):
        encoded_values = {}

        for feature, val in event.items():
            encoded_values.update(self._encode(feature, val))
        return encoded_values

    def _do_pandas(self, event):
        for key, values in self.mapping.items():
            event[key] = pd.Categorical(event[key], categories=list(values))
            encoded = pd.get_dummies(event[key], prefix=key, dtype=np.int64)
            col_rename = {
                name: OneHotEncoder._sanitized_category(name)
                for name in encoded.columns
            }
            encoded.rename(columns=col_rename, inplace=True)
            event = pd.concat([event.loc[:, :key], encoded, event.loc[:, key:]], axis=1)
        event.drop(columns=list(self.mapping.keys()), inplace=True)
        return event

    def _do_spark(self, event):
        from pyspark.sql.functions import lit, when

        for key, values in self.mapping.items():
            for val in values:
                event = event.withColumn(
                    f"{key}_{self._sanitized_category(val)}",
                    when(
                        (event[key] == val),
                        lit(1),
                    ).otherwise(lit(0)),
                )
        event = event.drop(*self.mapping.keys())
        return event

    @staticmethod
    def _sanitized_category(category):
        # replace(" " and "-") -> "_"
        if isinstance(category, str):
            return re.sub("[ -]", "_", category)
        return category


class DateExtractor(StepToDict, MLRunStep):
    def __init__(
        self,
        parts: Union[dict[str, str], list[str]],
        timestamp_col: Optional[str] = None,
        **kwargs,
    ):
        """Date Extractor extracts a date-time component into new columns

        The extracted date part will appear as `<timestamp_col>_<date_part>` feature.

        Supports part values:

        * asm8:              Return numpy datetime64 format in nanoseconds.
        * day_of_week:       Return day of the week.
        * day_of_year:       Return the day of the year.
        * dayofweek:         Return day of the week.
        * dayofyear:         Return the day of the year.
        * days_in_month:     Return the number of days in the month.
        * daysinmonth:       Return the number of days in the month.
        * freqstr:           Return the total number of days in the month.
        * is_leap_year:      Return True if year is a leap year.
        * is_month_end:      Return True if date is last day of month.
        * is_month_start:    Return True if date is first day of month.
        * is_quarter_end:    Return True if date is last day of the quarter.
        * is_quarter_start:  Return True if date is first day of the quarter.
        * is_year_end:       Return True if date is last day of the year.
        * is_year_start:     Return True if date is first day of the year.
        * quarter:           Return the quarter of the year.
        * tz:                Alias for tzinfo.
        * week:              Return the week number of the year.
        * weekofyear:        Return the week number of the year.

        example::

            # (taken from the fraud-detection end-to-end feature store demo)
            # Define the Transactions FeatureSet
            transaction_set = fstore.FeatureSet(
                "transactions",
                entities=[fstore.Entity("source")],
                timestamp_key="timestamp",
                description="transactions feature set",
            )

            # Get FeatureSet computation graph
            transaction_graph = transaction_set.graph

            # Add the custom `DateExtractor` step
            # to the computation graph
            transaction_graph.to(
                class_name="DateExtractor",
                name="Extract Dates",
                parts=["hour", "day_of_week"],
                timestamp_col="timestamp",
            )

        :param parts: list of pandas style date-time parts you want to extract.
        :param timestamp_col: The name of the column containing the timestamps to extract from,
                              by default "timestamp"
        """
        super().__init__(**kwargs)
        self.timestamp_col = timestamp_col if timestamp_col else "timestamp"
        self.parts = parts

    def _get_key_name(self, part: str):
        return f"{self.timestamp_col}_{part}"

    def _extract_timestamp(self, event):
        # Extract timestamp
        try:
            timestamp = event[self.timestamp_col]
        except KeyError:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"{self.timestamp_col} does not exist in the event"
            )
        return timestamp

    def _do_storey(self, event):
        timestamp = self._extract_timestamp(event)
        # Extract specified parts
        timestamp = pd.Timestamp(timestamp)
        for part in self.parts:
            # Extract part
            extracted_part = getattr(timestamp, part)
            # Add to event
            event[self._get_key_name(part)] = extracted_part
        return event

    def _do_pandas(self, event):
        timestamp = self._extract_timestamp(event)
        # Extract specified parts
        for part in self.parts:
            # Extract part and add it to event
            event[self._get_key_name(part)] = timestamp.map(
                lambda x: getattr(pd.Timestamp(x), part)
            )
        return event

    def _do_spark(self, event):
        import pyspark.sql.functions

        for part in self.parts:
            func = part
            # spark's naming for these functions is without underscores
            if func in ("day_of_year", "day_of_month"):
                func = func.replace("_", "")
            func = getattr(pyspark.sql.functions, func, None)
            if func:
                event = event.withColumn(
                    self._get_key_name(part),
                    func(self.timestamp_col).cast("long"),
                )
            else:
                raise mlrun.errors.MLRunRuntimeError(
                    f"DateExtractor with spark engine doesn't support {part} param"
                )
        return event


class SetEventMetadata(MapClass):
    def __init__(
        self,
        id_path: Optional[str] = None,
        key_path: Optional[str] = None,
        random_id: Optional[bool] = None,
        **kwargs,
    ) -> None:
        """Set the event metadata (id, key) from the event body

        set the event metadata fields (id and key) from the event body data structure
        the xx_path attribute defines the key or path to the value in the body dict, "." in the path string
        indicate the value is in a nested dict e.g. `"x.y"` means `{"x": {"y": value}}`

        example::

            flow = function.set_topology("flow")
            # build a graph and use the SetEventMetadata step to extract the id, key and path from the event body
            # ("myid" and "mykey" fields), the metadata will be used for following data processing steps
            # (e.g. feature store ops, key aggregations, write to databases/streams, etc.)
            flow.to(SetEventMetadata(id_path="myid", key_path="mykey"))
                .to(...)  # additional steps

            server = function.to_mock_server()
            event = {"myid": "34", "mykey": "123"}
            resp = server.test(body=event)

        :param id_path:   path to the id value
        :param key_path:  path to the key value
        :param random_id: if True will set the event.id to a random value
        """
        kwargs["full_event"] = True
        super().__init__(**kwargs)
        self.id_path = id_path
        self.key_path = key_path
        self.random_id = random_id

        self._tagging_funcs = []

    def post_init(self, mode="sync", **kwargs):
        def add_metadata(name, path, operator=str):
            def _add_meta(event):
                value = get_in(event.body, path)
                setattr(event, name, operator(value))

            return _add_meta

        def set_random_id(event):
            event.id = uuid.uuid4().hex

        self._tagging_funcs = []
        if self.id_path:
            self._tagging_funcs.append(add_metadata("id", self.id_path))
        if self.key_path:
            self._tagging_funcs.append(add_metadata("key", self.key_path))
        if self.random_id:
            self._tagging_funcs.append(set_random_id)

    def do(self, event):
        for func in self._tagging_funcs:
            func(event)
        return event


class DropFeatures(StepToDict, MLRunStep):
    def __init__(self, features: list[str], **kwargs):
        """Drop all the features from feature list

        :param features:    string list of the features names to drop

        example::

            feature_set = fstore.FeatureSet(
                "fs-new",
                entities=[fstore.Entity("id")],
                description="feature set",
                engine="pandas",
            )
            # Pre-processing graph steps
            feature_set.graph.to(DropFeatures(features=["age"]))
            df_pandas = feature_set.ingest(data)

        """
        super().__init__(**kwargs)
        self.features = features

    def _do_storey(self, event):
        for feature in self.features:
            try:
                del event[feature]
            except KeyError:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"The ingesting data doesn't contain a feature named '{feature}'"
                )
        return event

    def _do_pandas(self, event):
        return event.drop(columns=self.features)

    def _do_spark(self, event):
        return event.drop(*self.features)

    @classmethod
    def validate_args(cls, feature_set, **kwargs):
        features = kwargs.get("features", [])
        entity_names = list(feature_set.spec.entities.keys())
        dropped_entities = set(features).intersection(entity_names)
        if dropped_entities:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"DropFeatures can only drop features, not entities: {dropped_entities}"
            )
        if feature_set.spec.label_column in features:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"DropFeatures can not drop label_column: {feature_set.spec.label_column}"
            )
        if feature_set.spec.timestamp_key in features:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"DropFeatures can not drop timestamp_key: {feature_set.spec.timestamp_key}"
            )
